################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
################################################################################
import os
from typing import Any, Dict, List

import pytest
from chromadb.errors import NotFoundError

try:
    import chromadb  # noqa: F401

    chromadb_available = True
except ImportError:
    chromadb_available = False

from flink_agents.api.resource import Resource, ResourceType
from flink_agents.api.vector_stores.vector_store import (
    Document,
    VectorStoreQuery,
)
from flink_agents.integrations.vector_stores.chroma.chroma_vector_store import (
    ChromaVectorStore,
)

api_key = os.environ.get("TEST_API_KEY")
tenant = os.environ.get("TEST_TENANT")
database = os.environ.get("TEST_DATABASE")


class MockEmbeddingModel(Resource):  # noqa: D101
    name: str

    @classmethod
    def resource_type(cls) -> ResourceType:  # noqa: D102
        return ResourceType.EMBEDDING_MODEL

    @property
    def model_kwargs(self) -> Dict[str, Any]:  # noqa: D102
        return {}

    def embed(self, text: str, **kwargs: Any) -> list[float]:  # noqa: D102
        if "ChromaDB" in text:
            return [0.2, 0.3, 0.4, 0.5, 0.6]
        else:
            return [0.1, 0.2, 0.3, 0.4, 0.5]


def _populate_test_data(
    vector_store: ChromaVectorStore, collection_name: str | None = None
) -> List[Document]:
    """Private helper method to populate ChromaDB with test data."""
    vector_store.get_or_create_collection(name=collection_name)
    documents = [
        Document(
            id="doc1",
            content="ChromaDB is a vector database for AI applications",
            metadata={"category": "database", "source": "test"},
        ),
        Document(
            id="doc2",
            content="Apache Flink Agents is an Agentic AI framework based on Apache Flink.",
            metadata={"category": "ai-agent", "source": "test"},
        ),
    ]
    vector_store.add(documents=documents, collection_name=collection_name)
    return documents


@pytest.mark.skipif(not chromadb_available, reason="ChromaDB is not available")
def test_local_chroma_vector_store() -> None:
    """Test ChromaDB vector store with embedding model integration."""
    embedding_model = MockEmbeddingModel(name="mock_embeddings")

    def get_resource(name: str, resource_type: ResourceType) -> Resource:
        if resource_type == ResourceType.EMBEDDING_MODEL:
            return embedding_model
        else:
            msg = f"Unknown resource type: {resource_type}"
            raise ValueError(msg)

    vector_store = ChromaVectorStore(
        name="chroma_vector_store",
        embedding_model="mock_embeddings",
        collection="test_collection",
        get_resource=get_resource,
    )

    _populate_test_data(vector_store, "test_collection")

    query = VectorStoreQuery(query_text="What is Flink Agent?", limit=1)

    result = vector_store.query(query)
    assert result is not None
    assert len(result.documents) == 1
    assert result.documents[0].id == "doc2"


@pytest.mark.skipif(not chromadb_available, reason="ChromaDB is not available")
def test_collection_management() -> None:
    """Test ChromaDB vector store with embedding model integration."""
    embedding_model = MockEmbeddingModel(name="mock_embeddings")

    def get_resource(name: str, resource_type: ResourceType) -> Resource:
        if resource_type == ResourceType.EMBEDDING_MODEL:
            return embedding_model
        else:
            msg = f"Unknown resource type: {resource_type}"
            raise ValueError(msg)

    vector_store = ChromaVectorStore(
        name="chroma_vector_store",
        embedding_model="mock_embeddings",
        collection="test_collection",
        get_resource=get_resource,
    )

    vector_store.get_or_create_collection(
        name="collection_management", metadata={"key1": "value1", "key2": "value2"}
    )

    collection = vector_store.get_collection(name="collection_management")

    assert collection is not None
    assert collection.name == "collection_management"
    assert vector_store.size(collection_name="collection_management") == 0
    assert collection.metadata == {"key1": "value1", "key2": "value2"}

    vector_store.delete_collection(name="collection_management")

    with pytest.raises(NotFoundError):
        vector_store.get_collection(name="collection_management")


@pytest.mark.skipif(not chromadb_available, reason="ChromaDB is not available")
def test_document_management() -> None:
    """Test ChromaDB vector store with embedding model integration."""
    embedding_model = MockEmbeddingModel(name="mock_embeddings")

    def get_resource(name: str, resource_type: ResourceType) -> Resource:
        if resource_type == ResourceType.EMBEDDING_MODEL:
            return embedding_model
        else:
            msg = f"Unknown resource type: {resource_type}"
            raise ValueError(msg)

    vector_store = ChromaVectorStore(
        name="chroma_vector_store",
        embedding_model="mock_embeddings",
        collection="test_collection",
        get_resource=get_resource,
    )

    vector_store.get_or_create_collection(
        name="document_management", metadata={"key1": "value1", "key2": "value2"}
    )

    expected_documents = _populate_test_data(vector_store, "document_management")
    expected_documents[0].embedding = None
    expected_documents[1].embedding = None

    # test get all documents
    documents = vector_store.get(collection_name="document_management")
    assert documents == expected_documents

    # test get specific document
    documents = vector_store.get(ids="doc1", collection_name="document_management")
    assert len(documents) == 1
    assert documents[0] == expected_documents[0]

    # test delete specific document
    vector_store.delete(ids="doc1", collection_name="document_management")
    documents = vector_store.get(collection_name="document_management")
    assert len(documents) == 1
    assert documents[0] == expected_documents[1]

    # test delete all documents
    vector_store.delete(collection_name="document_management")
    documents = vector_store.get(collection_name="document_management")
    assert documents == []

    # test delete from empty collection
    vector_store.delete(collection_name="document_management")
    documents = vector_store.get(collection_name="document_management")
    assert documents == []


@pytest.mark.skipif(api_key is None, reason="TEST_API_KEY is not set")
def test_cloud_chroma_vector_store() -> None:
    """Test cloud ChromaDB vector store with embedding model integration."""
    embedding_model = MockEmbeddingModel(name="mock_embeddings")

    def get_resource(name: str, resource_type: ResourceType) -> Resource:
        if resource_type == ResourceType.EMBEDDING_MODEL:
            return embedding_model
        else:
            msg = f"Unknown resource type: {resource_type}"
            raise ValueError(msg)

    vector_store = ChromaVectorStore(
        name="chroma_vector_store",
        embedding_model="mock_embeddings",
        collection="test_collection",
        api_key=api_key,
        tenant=tenant,
        database=database,
        get_resource=get_resource,
    )

    _populate_test_data(vector_store)

    query = VectorStoreQuery(query_text="What is Flink Agent?", limit=1)

    result = vector_store.query(query)
    assert result is not None
    assert len(result.documents) == 1
    assert result.documents[0].id == "doc2"



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import json
from pathlib import Path

import pytest
from pyflink.common.typeinfo import BasicTypeInfo, RowTypeInfo

from flink_agents.api.agents.react_agent import OutputSchema
from flink_agents.api.events.event import InputEvent
from flink_agents.api.runner_context import RunnerContext
from flink_agents.plan.actions.action import Action
from flink_agents.plan.function import PythonFunction


def legal_signature(event: InputEvent, ctx: RunnerContext) -> None:  # noqa: D103
    pass


def illegal_signature(value: int, ctx: RunnerContext) -> None:  # noqa: D103
    pass


def test_action_signature_legal() -> None:  # noqa: D103
    Action(
        name="legal",
        exec=PythonFunction.from_callable(legal_signature),
        listen_event_types=[f"{InputEvent.__module__}.{InputEvent.__qualname__}"],
    )


def test_action_signature_illegal() -> None:  # noqa: D103
    with pytest.raises(TypeError):
        Action(
            name="illegal",
            exec=PythonFunction.from_callable(illegal_signature),
            listen_event_types=[f"{InputEvent.__module__}.{InputEvent.__qualname__}"],
        )


@pytest.fixture(scope="module")
def action() -> Action:  # noqa: D103
    func = PythonFunction.from_callable(legal_signature)
    return Action(
        name="legal",
        exec=func,
        listen_event_types=[f"{InputEvent.__module__}.{InputEvent.__qualname__}"],
        config={
            "output_schema": OutputSchema(
                output_schema=RowTypeInfo(
                    [BasicTypeInfo.INT_TYPE_INFO()],
                    ["result"],
                )
            )
        },
    )


current_dir = Path(__file__).parent


def test_action_serialize(action: Action) -> None:  # noqa: D103
    json_value = action.model_dump_json(serialize_as_any=True, indent=4)
    with Path.open(Path(f"{current_dir}/resources/action.json")) as f:
        expected_json = f.read()
    actual = json.loads(json_value)
    expected = json.loads(expected_json)
    assert actual == expected


def test_action_deserialize(action: Action) -> None:  # noqa: D103
    with Path.open(Path(f"{current_dir}/resources/action.json")) as f:
        expected_json = f.read()
    action = Action.model_validate_json(expected_json)
    assert action.name == "legal"
    assert action.listen_event_types == ["flink_agents.api.events.event.InputEvent"]
    func = action.exec
    assert func.module == "flink_agents.plan.tests.test_action"
    assert func.qualname == "legal_signature"



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import logging

from flink_agents.api.core_options import AgentExecutionOptions
from flink_agents.api.events.context_retrieval_event import (
    ContextRetrievalRequestEvent,
    ContextRetrievalResponseEvent,
)
from flink_agents.api.events.event import Event
from flink_agents.api.resource import ResourceType
from flink_agents.api.runner_context import RunnerContext
from flink_agents.api.vector_stores.java_vector_store import JavaVectorStore
from flink_agents.api.vector_stores.vector_store import VectorStoreQuery
from flink_agents.plan.actions.action import Action
from flink_agents.plan.function import PythonFunction

_logger = logging.getLogger(__name__)

async def process_context_retrieval_request(event: Event, ctx: RunnerContext) -> None:
    """Built-in action for processing context retrieval requests."""
    if isinstance(event, ContextRetrievalRequestEvent):
        vector_store = ctx.get_resource(event.vector_store, ResourceType.VECTOR_STORE)

        query = VectorStoreQuery(query_text=event.query, limit=event.max_results)

        rag_async = ctx.config.get(AgentExecutionOptions.RAG_ASYNC)
        # java vector store doesn't support async execution
        # see https://github.com/apache/flink-agents/issues/448 for details.
        rag_async = rag_async and not isinstance(vector_store, JavaVectorStore)
        if rag_async:
            # To avoid https://github.com/alibaba/pemja/issues/88,
            # we log a message here.
            _logger.debug("Processing context retrieval asynchronously.")
            result = await ctx.durable_execute_async(vector_store.query, query)
        else:
            result = ctx.durable_execute(vector_store.query, query)

        ctx.send_event(
            ContextRetrievalResponseEvent(
                request_id=event.id, query=event.query, documents=result.documents
            )
        )


CONTEXT_RETRIEVAL_ACTION = Action(
    name="context_retrieval_action",
    exec=PythonFunction.from_callable(process_context_retrieval_request),
    listen_event_types=[
        f"{ContextRetrievalRequestEvent.__module__}.{ContextRetrievalRequestEvent.__name__}",
    ],
)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import importlib
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Type

from pydantic import (
    BaseModel,
    Field,
    field_serializer,
    model_validator,
)

from flink_agents.api.chat_message import ChatMessage
from flink_agents.api.configuration import ConfigOption
from flink_agents.api.prompts.prompt import Prompt

ItemType = str | ChatMessage

class CompactionConfig(BaseModel):
    """Compaction configuration.

    Attributes:
        model: The name of the llm used for generating the summarization.
        prompt: The prompt instance or  name of the prompt used for generating
            the summarization. Optional.
        limit: How many summarization should generate for the context. Defaults to 1.
    """

    model: str
    prompt: str | Prompt | None = None
    limit: int = 1

class LongTermMemoryBackend(Enum):
    """Backend for Long-Term Memory."""

    EXTERNAL_VECTOR_STORE = "external_vector_store"


class LongTermMemoryOptions:
    """Config options for ReActAgent."""

    BACKEND = ConfigOption(
        key="long-term-memory.backend",
        config_type=LongTermMemoryBackend,
        default=None,
    )

    EXTERNAL_VECTOR_STORE_NAME = ConfigOption(
        key="long-term-memory.external-vector-store-name",
        config_type=str,
        default=None,
    )

    ASYNC_COMPACTION = ConfigOption(
        key="long-term-memory.async-compaction",
        config_type=bool,
        default=True,
    )


class DatetimeRange(BaseModel):
    """Represents a datetime range."""

    start: datetime
    end: datetime


class MemorySetItem(BaseModel):
    """Represents a long term memory item retrieved from vector store.

    Attributes:
        memory_set_name: The name of the memory set this item belongs to.
        id: The id of this item.
        value: The value of this item.
        compacted: Whether this item has been compacted.
        created_time: The timestamp this item was added to the memory set.
        last_accessed_time: The timestamp this item was last accessed.
        additional_metadata: Additional metadata for this item.
    """

    memory_set_name: str
    id: str
    value: Any
    compacted: bool = False
    created_time: datetime | DatetimeRange = None
    last_accessed_time: datetime
    additional_metadata: Dict[str, Any] | None = None


class MemorySet(BaseModel):
    """Represents a long term memory set contains memory items.

    Attributes:
        name: The name of this memory set.
        item_type: The type of items stored in this set.
        capacity: The capacity of this memory set.
        compaction_config: Compaction config
        to compact memory set.
    """

    name: str
    item_type: Type[str] | Type[ChatMessage]
    capacity: int
    compaction_config: CompactionConfig
    ltm: "BaseLongTermMemory" = Field(default=None, exclude=True)

    @field_serializer("item_type")
    def _serialize_item_type(self, item_type: Type) -> Dict[str, str]:
        return {"module": item_type.__module__, "name": item_type.__name__}

    @model_validator(mode="before")
    def _deserialize_item_type(self) -> "MemorySet":
        if isinstance(self["item_type"], Dict):
            module = importlib.import_module(self["item_type"]["module"])
            self["item_type"] = getattr(module, self["item_type"]["name"])
        return self

    @property
    def size(self) -> int:
        """The count of items of this memory set."""
        return self.ltm.size(memory_set=self)

    def add(
        self, items: ItemType | List[ItemType], ids: str | List[str] | None = None
    ) -> List[str]:
        """Add a memory item to the set, currently only support item with
        type str or ChatMessage.

        If the capacity of this memory set is reached, will trigger reduce
        operation to manage the memory set size.

        Args:
            items: The items to be inserted to this set.
            ids: The ids of the items to be inserted. Optional.

        Returns:
            The IDs of the items added.
        """
        return self.ltm.add(memory_set=self, memory_items=items, ids=ids)

    def get(
        self, ids: str | List[str] | None = None
    ) -> MemorySetItem | List[MemorySetItem]:
        """Retrieve memory items. If no item id provided, will return all items.

        Args:
            ids: The ids of the items to retrieve.

        Returns:
            The memory items retrieved.
        """
        return self.ltm.get(memory_set=self, ids=ids)

    def search(self, query: str, limit: int, **kwargs: Any) -> List[MemorySetItem]:
        """Retrieve n memory items related to the query.

        Args:
            query: The query to search for.
            limit: The number of items to retrieve.
            **kwargs: Additional arguments for search.
        """
        return self.ltm.search(memory_set=self, query=query, limit=limit, **kwargs)


class BaseLongTermMemory(ABC, BaseModel):
    """Base Abstract class for long term memory."""

    @abstractmethod
    def get_or_create_memory_set(
        self,
        name: str,
        item_type: type[str] | Type[ChatMessage],
        capacity: int,
        compaction_config: CompactionConfig,
    ) -> MemorySet:
        """Create a memory set, if the memory set already exists, return it.

        Args:
            name: The name of the memory set.
            item_type: The type of the memory item.
            capacity: The capacity of the memory set.
            compaction_config: The compaction config
            storge management.

        Returns:
            The created memory set.
        """

    @abstractmethod
    def get_memory_set(self, name: str) -> MemorySet:
        """Get the memory set.

        Args:
            name: The name of the memory set.

        Returns:
            The memory set.
        """

    @abstractmethod
    def delete_memory_set(self, name: str) -> bool:
        """Delete the memory set.

        Args:
            name: The name of the memory set.

        Returns:
            Whether the memory set was deleted.
        """

    @abstractmethod
    def size(self, memory_set: MemorySet) -> int:
        """Get the size of the memory set.

        Args:
            memory_set: The memory set to count.
        """

    @abstractmethod
    def add(
        self,
        memory_set: MemorySet,
        memory_items: ItemType | List[ItemType],
        ids: str | List[str] | None = None,
        metadatas: Dict[str, Any] | List[Dict[str, Any]] | None = None,
    ) -> List[str]:
        """Add items to the memory set, currently only support items with
        type str or ChatMessage.

        This method may trigger compaction to manage the memory set size.

        Args:
            memory_set: The memory set to add to.
            memory_items: The items to be added to this set.
            ids: The IDs of items. Will be automatically generated if not provided.
            Optional.
            metadatas: The metadata for items. Optional.

        Returns:
            The IDs of added items.
        """

    @abstractmethod
    def get(
        self, memory_set: MemorySet, ids: str | List[str] | None = None
    ) -> MemorySetItem | List[MemorySetItem]:
        """Retrieve memory items. If no item id provided, return all items.

        Args:
            memory_set: The set to be retrieved.
            ids: The ids of the items to retrieve. If not provided, all items will
            be retrieved. Optional.

        Returns:
            The memory items retrieved.
        """

    @abstractmethod
    def delete(self, memory_set: MemorySet, ids: str | List[str] | None = None) -> None:
        """Delete memory items. If no item id provided, delete all items.

        Args:
            memory_set: The memory set to delete from.
            ids: The ids of items to be deleted, If not provided, all items will be
            deleted. Optional.
        """

    @abstractmethod
    def search(
        self, memory_set: MemorySet, query: str, limit: int, **kwargs: Any
    ) -> List[MemorySetItem]:
        """Retrieve n memory items related to the query.

        Args:
            memory_set: The set to be retrieved.
            query: The query for sematic search.
            limit: The number of items to retrieve.
            **kwargs: Additional arguments for sematic search.

        Returns:
            Related memory items retrieved.
        """

    @abstractmethod
    def close(self) -> None:
        """Logic executed when job close."""



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import os
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List
from unittest.mock import create_autospec

import pytest
from chromadb import EmbeddingFunction
from chromadb.utils import embedding_functions
from pydantic import ConfigDict

from flink_agents.api.chat_message import ChatMessage, MessageRole
from flink_agents.api.memory.long_term_memory import (
    CompactionConfig,
    DatetimeRange,
    MemorySet,
)
from flink_agents.api.metric_group import MetricGroup
from flink_agents.api.resource import Resource, ResourceType
from flink_agents.api.runner_context import RunnerContext
from flink_agents.integrations.chat_models.ollama_chat_model import (
    OllamaChatModelConnection,
    OllamaChatModelSetup,
)
from flink_agents.integrations.embedding_models.local.ollama_embedding_model import (
    OllamaEmbeddingModelConnection,
    OllamaEmbeddingModelSetup,
)
from flink_agents.integrations.vector_stores.chroma.chroma_vector_store import (
    ChromaVectorStore,
)
from flink_agents.runtime.memory.vector_store_long_term_memory import (
    VectorStoreLongTermMemory,
)

if TYPE_CHECKING:
    from flink_agents.api.memory.long_term_memory import (
        MemorySetItem,
    )

current_dir = Path(__file__).parent


class MockEmbeddingModel(Resource):  # noqa: D101
    model_config = ConfigDict(arbitrary_types_allowed=True)
    ef: EmbeddingFunction = embedding_functions.DefaultEmbeddingFunction()

    @classmethod
    def resource_type(cls) -> ResourceType:  # noqa: D102
        return ResourceType.EMBEDDING_MODEL

    @property
    def model_kwargs(self) -> Dict[str, Any]:  # noqa: D102
        return {}

    def embed(self, text: str, **kwargs: Any) -> Any:  # noqa: D102
        return self.ef([text])[0]


@pytest.fixture(scope="module")
def long_term_memory() -> VectorStoreLongTermMemory:  # noqa: D103
    embedding_model_connection = OllamaEmbeddingModelConnection()

    chat_model_connection = OllamaChatModelConnection(request_timeout=240)

    use_ollama = os.environ.get("USE_OLLAMA")

    def get_resource(name: str, type: ResourceType) -> Resource:
        if type == ResourceType.CHAT_MODEL:
            return chat_model
        elif type == ResourceType.CHAT_MODEL_CONNECTION:
            return chat_model_connection
        elif type == ResourceType.EMBEDDING_MODEL:
            if use_ollama:
                return embedding_model
            else:
                return MockEmbeddingModel()
        elif type == ResourceType.EMBEDDING_MODEL_CONNECTION:
            return embedding_model_connection
        else:
            return vector_store

    chat_model = OllamaChatModelSetup(
        get_resource=get_resource, connection="chat_model_connection", model="qwen3:8b"
    )

    embedding_model = OllamaEmbeddingModelSetup(
        get_resource=get_resource,
        connection="embedding_model_connection",
        model="nomic-embed-text",
    )

    chromadb_path = tempfile.mkdtemp()
    print(f"Starting ChromaDB in {chromadb_path}...")

    vector_store = ChromaVectorStore(
        persist_directory=chromadb_path,
        embedding_model="embedding_model",
        get_resource=get_resource,
    )

    mock_runner_context = create_autospec(RunnerContext, instance=True)
    mock_runner_context.get_resource = get_resource
    mock_runner_context.agent_metric_group.get_sub_group.return_value = create_autospec(
        MetricGroup, instance=True
    )

    return VectorStoreLongTermMemory(
        ctx=mock_runner_context,
        vector_store="vector_store",
        job_id="bc0b2ad61ecd4a615d92ce25390f61ad",
        key="00001",
    )


def prepare_memory_set(  # noqa: D103
    long_term_memory: VectorStoreLongTermMemory,
    compaction_config: CompactionConfig = CompactionConfig(model="llm"),  # noqa:B008
) -> (MemorySet, List[ChatMessage]):
    memory_set: MemorySet = long_term_memory.get_or_create_memory_set(
        name="chat_history",
        item_type=ChatMessage,
        capacity=100,
        compaction_config=compaction_config,
    )

    msgs: List[ChatMessage] = []
    for i in range(20):
        msg = ChatMessage(role=MessageRole.USER, content=f"This is the no.{i} message.")
        msgs.append(msg)

    memory_set.add(items=msgs)
    return memory_set, msgs


def test_get_memory_set(  # noqa:D103
    long_term_memory: VectorStoreLongTermMemory,
) -> None:
    memory_set, _ = prepare_memory_set(long_term_memory)
    retrieved = long_term_memory.get_memory_set(memory_set.name)
    assert retrieved == memory_set

    long_term_memory.delete_memory_set(name="chat_history")


def test_add_and_get(  # noqa:D103
    long_term_memory: VectorStoreLongTermMemory,
) -> None:
    memory_set, msgs = prepare_memory_set(long_term_memory)

    retrieved: List[MemorySetItem] = memory_set.get()
    retrieved_msgs = [item.value for item in retrieved]

    assert retrieved_msgs == msgs

    long_term_memory.delete_memory_set(name="chat_history")


def test_search(  # noqa:D103
    long_term_memory: VectorStoreLongTermMemory,
) -> None:
    memory_set, msgs = prepare_memory_set(long_term_memory)

    retrieved: List[MemorySetItem] = memory_set.search(
        query="The no.10 message", limit=1
    )
    retrieved_msgs = [item.value for item in retrieved]

    assert retrieved_msgs == msgs[10:11]

    long_term_memory.delete_memory_set(name="chat_history")


@pytest.mark.skip("Depend on ollama server")
def test_compact(  # noqa:D103
    long_term_memory: VectorStoreLongTermMemory,
) -> None:
    memory_set: MemorySet = long_term_memory.get_or_create_memory_set(
        name="chat_history",
        item_type=ChatMessage,
        capacity=8,
        compaction_config=CompactionConfig(model="llm", limit=2),
    )

    msgs: List[ChatMessage] = [
        ChatMessage(role=MessageRole.USER, content="What is flink?"),
        ChatMessage(
            role=MessageRole.ASSISTANT,
            content="Apache Flink is a framework and distributed processing engine for stateful computations over unbounded and bounded data streams. Flink has been designed to run in all common cluster environments, perform computations at in-memory speed and at any scale.",
        ),
        ChatMessage(role=MessageRole.USER, content="What is flink agents?"),
        ChatMessage(
            role=MessageRole.ASSISTANT,
            content="Apache Flink Agents is a brand-new sub-project from the Apache Flink community, providing an open-source framework for building event-driven streaming agents.",
        ),
        ChatMessage(
            role=MessageRole.USER, content="What's the whether tomorrow in london?"
        ),
        ChatMessage(
            role=MessageRole.ASSISTANT,
            content="",
            tool_calls=[
                {
                    "id": "186780f8-c79d-4159-83e3-f65859835b14",
                    "type": "function",
                    "function": {
                        "name": "get_weather",
                        "arguments": {"position": "london", "time": "tomorrow"},
                    },
                }
            ],
        ),
        ChatMessage(role=MessageRole.TOOL, content="snow"),
        ChatMessage(
            role=MessageRole.ASSISTANT, content="Tomorrow weather for london is snow."
        ),
    ]
    memory_set.add(items=msgs)

    retrieved: List[MemorySetItem] = memory_set.get()
    retrieved_msgs = [item.value for item in retrieved]

    """
    The retrieved messages likes:
        [ChatMessage(role= < MessageRole.USER: 'user' >, content = 'User inquire about
            Apache Flink and Flink Agents, and the assistant explains Flink as a stream
            processing framework and Flink Agents as its sub-project.',
            tool_calls = [], extra_args = {}),
        ChatMessage(role= < MessageRole.USER: 'user' >, content = "User ask for London's
            weather tomorrow, llm invoke get_weather function, and the answer is snow.",
            tool_calls = [], extra_args = {})]
    The summarization quality is not stable on qwen3:8b.
    """

    assert len(retrieved_msgs) == 2
    assert retrieved[0].compacted
    assert retrieved[1].compacted
    assert isinstance(retrieved[0].created_time, DatetimeRange)
    assert isinstance(retrieved[1].created_time, DatetimeRange)
    assert memory_set.size == 2

    long_term_memory.delete_memory_set(name="chat_history")



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
from typing import Any, Dict, List
from uuid import UUID

from flink_agents.api.events.event import Event


class ToolRequestEvent(Event):
    """Event representing a tool call request.

    Attributes:
    ----------
    model: str
        name of the model that generated the tool request.
    tool_calls : List[Dict[str, Any]]
        tool calls that should be executed in batch.
    """

    model: str
    tool_calls: List[Dict[str, Any]]


class ToolResponseEvent(Event):
    """Event representing a result from tool call.

    Attributes:
    ----------
    request_id : UUID
        The id of the request event.
    responses : Dict[UUID, Any]
        The dict maps tool call id to result.
    external_ids : Dict[UUID, str]
        Optional identifier for storing original tool call IDs from external systems
        (e.g., Anthropic tool_use_id).
    """

    request_id: UUID
    responses: Dict[UUID, Any]
    external_ids: Dict[UUID, str | None]



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
from typing import Callable, Type

from flink_agents.api.events.event import Event


def action(*listen_events: Type[Event]) -> Callable:
    """Decorator for marking a function as an agent action.

    Parameters
    ----------
    listen_events : list[Type[Event]]
        List of event types that this action should respond to.

    Returns:
    -------
    Callable
        Decorator function that marks the target function with event listeners.

    Raises:
    ------
    AssertionError
        If no events are provided to listen to.
    """
    assert len(listen_events) > 0, (
        "action must have at least one event type to listen to"
    )

    for event in listen_events:
        assert issubclass(event, Event), "action must only listen to event types."

    def decorator(func: Callable) -> Callable:
        func._listen_events = listen_events
        return func

    return decorator


def chat_model_connection(func: Callable) -> Callable:
    """Decorator for marking a function declaring a chat model connection.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns:
    -------
    Callable
        Decorator function that marks the target function declare a chat model
        connection.
    """
    func._is_chat_model_connection = True
    return func


def chat_model_setup(func: Callable) -> Callable:
    """Decorator for marking a function declaring a chat model setup.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns:
    -------
    Callable
        Decorator function that marks the target function declare a chat model.
    """
    func._is_chat_model_setup = True
    return func


def embedding_model_connection(func: Callable) -> Callable:
    """Decorator for marking a function declaring an embedding model connection.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns:
    -------
    Callable
        Decorator function that marks the target function declare an embedding model
        connection.
    """
    func._is_embedding_model_connection = True
    return func


def embedding_model_setup(func: Callable) -> Callable:
    """Decorator for marking a function declaring an embedding model setup.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns:
    -------
    Callable
        Decorator function that marks the target function declare an embedding model.
    """
    func._is_embedding_model_setup = True
    return func


def tool(func: Callable) -> Callable:
    """Decorator for marking a function declaring a tool.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns:
    -------
    Callable
        Decorator function that marks the target function declare a tool.
    """
    func._is_tool = True
    return func


def prompt(func: Callable) -> Callable:
    """Decorator for marking a function declaring a prompt.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns:
    -------
    Callable
        Decorator function that marks the target function declare a prompt.
    """
    func._is_prompt = True
    return func


def mcp_server(func: Callable) -> Callable:
    """Decorator for marking a function declaring a MCP server.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns:
    -------
    Callable
        Decorator function that marks the target function declare a MCP server.
    """
    func._is_mcp_server = True
    return func


def vector_store(func: Callable) -> Callable:
    """Decorator for marking a function declaring a vector store.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns:
    -------
    Callable
        Decorator function that marks the target function declare a vector store.
    """
    func._is_vector_store = True
    return func

def java_resource(cls: Type) -> Type:
    """Decorator to mark a class as Java resource."""
    cls._is_java_resource = True
    return cls



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import time

import pytest

from flink_agents.api.agents.agent import Agent
from flink_agents.api.decorators import action
from flink_agents.api.events.event import Event, InputEvent, OutputEvent
from flink_agents.api.execution_environment import AgentsExecutionEnvironment
from flink_agents.api.runner_context import RunnerContext


class Agent1(Agent):  # noqa: D101
    @action(InputEvent)
    @staticmethod
    def increment(event: Event, ctx: RunnerContext):  # noqa D102
        input = event.input
        value = input + 1
        ctx.send_event(OutputEvent(output=value))


class Agent1WithAsync(Agent):  # noqa: D101
    @action(InputEvent)
    @staticmethod
    async def increment(event: Event, ctx: RunnerContext):  # noqa D102
        def my_func(value: int) -> int:
            time.sleep(1)
            return value + 1

        input = event.input
        value = await ctx.durable_execute_async(my_func, input)
        ctx.send_event(OutputEvent(output=value))


class Agent2(Agent):  # noqa: D101
    @action(InputEvent)
    @staticmethod
    def decrease(event: Event, ctx: RunnerContext):  # noqa D102
        input = event.input
        value = input - 1
        ctx.send_event(OutputEvent(output=value))


def test_local_execution_environment() -> None:  # noqa: D103
    env = AgentsExecutionEnvironment.get_execution_environment()

    input_list = []
    agent = Agent1()

    output_list = env.from_list(input_list).apply(agent).to_list()

    input_list.append({"key": "bob", "value": 1})
    input_list.append({"k": "john", "v": 2})

    env.execute()

    assert output_list == [{"bob": 2}, {"john": 3}]


def test_local_execution_environment_with_async() -> None:  # noqa: D103
    env = AgentsExecutionEnvironment.get_execution_environment()

    input_list = []
    agent = Agent1WithAsync()

    output_list = env.from_list(input_list).apply(agent).to_list()

    input_list.append({"key": "bob", "value": 1})
    input_list.append({"k": "john", "v": 2})

    env.execute()

    assert output_list == [{"bob": 2}, {"john": 3}]


def test_local_execution_environment_apply_multi_agents() -> None:  # noqa: D103
    env = AgentsExecutionEnvironment.get_execution_environment()

    input_list = []
    agent1 = Agent1()
    agent2 = Agent2()

    with pytest.raises(RuntimeError):
        env.from_list(input_list).apply(agent1).apply(agent2).to_list()


def test_local_execution_environment_execute_multi_times() -> None:  # noqa: D103
    env = AgentsExecutionEnvironment.get_execution_environment()

    input_list = []
    agent = Agent1()

    env.from_list(input_list).apply(agent).to_list()

    input_list.append({"key": "bob", "value": 1})
    input_list.append({"k": "john", "v": 2})

    env.execute()
    with pytest.raises(RuntimeError):
        env.execute()


def test_local_execution_environment_call_from_list_twice() -> None:  # noqa: D103
    env = AgentsExecutionEnvironment.get_execution_environment()

    input_list = []

    env.from_list(input_list)
    with pytest.raises(RuntimeError):
        env.from_list(input_list)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################

from flink_agents.api.chat_message import ChatMessage
from flink_agents.api.memory.long_term_memory import (
    CompactionConfig,
    MemorySet,
)


def test_memory_set_serialization() -> None:  # noqa:D103
    memory_set = MemorySet(
        name="chat_history",
        item_type=ChatMessage,
        capacity=100,
        compaction_config=CompactionConfig(model="llm"),
    )

    json_data = memory_set.model_dump_json()

    memory_set_deserialized = MemorySet.model_validate_json(json_data)

    assert memory_set_deserialized == memory_set



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import uuid
from typing import Any, Dict, Generator, List

import chromadb
from chromadb import ClientAPI as ChromaClient
from chromadb import CloudClient
from chromadb.config import Settings
from pydantic import Field
from typing_extensions import override

from flink_agents.api.vector_stores.vector_store import (
    Collection,
    CollectionManageableVectorStore,
    Document,
)

DEFAULT_COLLECTION = "flink_agents_chroma_collection"

MAX_CHUNK_SIZE = 41665


def chunk_list(
    lst: List[Document], max_chunk_size: int
) -> Generator[List[Document], None, None]:
    """Yield successive max_chunk_size-sized chunks from list.

    This method is inspired by `chunk_list` in llama_index

    Args:
        lst : list of documents with embeddings
        max_chunk_size : max chunk size

    Yields:
        Generator[List[Document], None, None]: list of documents with embeddings

    """
    for i in range(0, len(lst), max_chunk_size):
        yield lst[i : i + max_chunk_size]


class ChromaVectorStore(CollectionManageableVectorStore):
    """ChromaDB vector store that handles connection and semantic search.

    Visit https://docs.trychroma.com/ for ChromaDB documentation.

    Supports multiple client modes:
    - In-memory: No configuration needed (default)
    - Persistent: Provide persist_directory
    - Server: Provide host and port
    - Cloud: Provide api_key for Chroma Cloud

    Attributes:
    ----------
    persist_directory : Optional[str]
        Directory for persistent storage. If None, uses in-memory client.
    host : Optional[str]
        Host for ChromaDB server connection.
    port : Optional[int]
        Port for ChromaDB server connection.
    api_key : Optional[str]
        API key for Chroma Cloud connection.
    client_settings : Optional[Settings]
        ChromaDB client settings for advanced configuration.
    tenant : str
        ChromaDB tenant for multi-tenancy support (default: "default_tenant").
    database : str
        ChromaDB database name (default: "default_database").
    collection : str
        Name of the ChromaDB collection to use (default: flink_agents_collection).
    collection_metadata : Dict[str, Any]
        Metadata for the collection (optional).
    create_collection_if_not_exists : bool
        Whether to create the collection if it doesn't exist (default: True).
    """

    # Connection configuration
    persist_directory: str | None = Field(
        default=None,
        description="Directory for persistent storage. If None, uses in-memory client.",
    )
    host: str | None = Field(
        default=None,
        description="Host for ChromaDB server connection.",
    )
    port: int | None = Field(
        default=8000,
        description="Port for ChromaDB server connection.",
    )
    api_key: str | None = Field(
        default=None,
        description="API key for Chroma Cloud connection.",
    )
    client_settings: Settings | None = Field(
        default=None,
        description="ChromaDB client settings for advanced configuration.",
    )
    tenant: str = Field(
        default="default_tenant",
        description="ChromaDB tenant for multi-tenancy support.",
    )
    database: str = Field(
        default="default_database",
        description="ChromaDB database name.",
    )

    # Collection configuration
    collection: str = Field(
        default=DEFAULT_COLLECTION,
        description="Name of the ChromaDB collection to use.",
    )
    collection_metadata: Dict[str, Any] = Field(
        default_factory=dict,
        description="Metadata for the collection.",
    )
    create_collection_if_not_exists: bool = Field(
        default=True,
        description="Whether to create the collection if it doesn't exist.",
    )

    __client: ChromaClient | None = None

    def __init__(
        self,
        *,
        embedding_model: str,
        persist_directory: str | None = None,
        host: str | None = None,
        port: int | None = 8000,
        api_key: str | None = None,
        client_settings: Settings | None = None,
        tenant: str = "default_tenant",
        database: str = "default_database",
        collection: str = DEFAULT_COLLECTION,
        collection_metadata: Dict[str, Any] | None = None,
        create_collection_if_not_exists: bool = True,
        **kwargs: Any,
    ) -> None:
        """Init method."""
        if collection_metadata is None:
            collection_metadata = {}
        super().__init__(
            embedding_model=embedding_model,
            persist_directory=persist_directory,
            host=host,
            port=port,
            api_key=api_key,
            client_settings=client_settings,
            tenant=tenant,
            database=database,
            collection=collection,
            collection_metadata=collection_metadata,
            create_collection_if_not_exists=create_collection_if_not_exists,
            **kwargs,
        )

    @property
    def client(self) -> ChromaClient:
        """Return ChromaDB client, creating it if necessary."""
        if self.__client is None:
            # Choose client type based on configuration
            if self.api_key is not None:
                # Cloud mode
                self.__client = CloudClient(
                    tenant=self.tenant,
                    database=self.database,
                    api_key=self.api_key,
                )
            elif self.host is not None:
                # Client-Server Mode
                self.__client = chromadb.HttpClient(
                    host=self.host,
                    port=self.port,
                    settings=self.client_settings,
                    tenant=self.tenant,
                    database=self.database,
                )
            elif self.persist_directory is not None:
                # Persistent mode
                self.__client = chromadb.PersistentClient(
                    path=self.persist_directory,
                    settings=self.client_settings,
                    tenant=self.tenant,
                    database=self.database,
                )
            else:
                # In-memory mode
                self.__client = chromadb.EphemeralClient(
                    settings=self.client_settings,
                    tenant=self.tenant,
                    database=self.database,
                )

        return self.__client

    @property
    def store_kwargs(self) -> Dict[str, Any]:
        """Return ChromaDB-specific setup settings."""
        return {
            "collection": self.collection,
            "collection_metadata": self.collection_metadata,
            "create_collection_if_not_exists": self.create_collection_if_not_exists,
        }

    @override
    def get_or_create_collection(
        self, name: str, metadata: Dict[str, Any] | None = None
    ) -> Collection:
        collection = self.client.get_or_create_collection(name=name, metadata=metadata)
        return Collection(name=collection.name, metadata=collection.metadata)

    @override
    def get_collection(self, name: str) -> Collection:
        collection = self.client.get_collection(name=name)
        return Collection(name=collection.name, metadata=collection.metadata)

    @override
    def delete_collection(self, name: str) -> Collection:
        collection = self.get_collection(name=name)
        self.client.delete_collection(name=collection.name)
        return collection

    @override
    def size(self, collection_name: str | None = None) -> int:
        return self.client.get_collection(
            name=collection_name or self.collection
        ).count()

    @override
    def get(
        self,
        ids: str | List[str] | None = None,
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> List[Document]:
        where = kwargs.get("where")
        limit = kwargs.get("limit")
        offset = kwargs.get("offset")
        where_document = kwargs.get("where_document")
        results = self.client.get_collection(
            name=collection_name or self.collection
        ).get(
            ids=ids,
            where=where,
            limit=limit,
            offset=offset,
            where_document=where_document,
        )

        ids = results["ids"]
        documents = results["documents"]
        metadatas = results["metadatas"]

        return [
            Document(id=id, content=document, metadata=dict(metadata))
            for id, document, metadata in zip(ids, documents, metadatas, strict=False)
        ]

    @override
    def delete(
        self,
        ids: str | List[str] | None = None,
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> None:
        where = kwargs.get("where")
        where_document = kwargs.get("where_document")
        if ids is None and where is None and where_document is None:
            ids = (
                self.client.get_collection(collection_name or self.collection)
                .get(include=[])
                .get("ids")
            )
            # collection is empty
            if len(ids) == 0:
                return
        self.client.get_collection(name=collection_name or self.collection).delete(
            ids=ids, where=where, where_document=where_document
        )

    @override
    def _add_embedding(
        self,
        *,
        documents: List[Document],
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> List[str]:
        collection_name = collection_name or self.collection
        collection_metadata = kwargs.get(
            "collection_metadata", self.collection_metadata
        )
        create_collection_if_not_exists = kwargs.get(
            "create_collection_if_not_exists", self.create_collection_if_not_exists
        )

        # Get or create collection based on configuration
        if create_collection_if_not_exists:
            # ChromaDB doesn't accept empty metadata, pass None instead
            metadata = collection_metadata if collection_metadata else None
            collection = self.client.get_or_create_collection(
                name=collection_name,
                metadata=metadata,
            )
        else:
            collection = self.client.get_collection(name=collection_name)

        documents_chunks = chunk_list(documents, MAX_CHUNK_SIZE)

        all_ids = []
        for chunk in documents_chunks:
            ids = [doc.id or str(uuid.uuid4()) for doc in chunk]
            docs = [doc.content for doc in chunk]
            embeddings = [doc.embedding for doc in chunk]
            metadatas = [doc.metadata for doc in chunk]

            collection.add(
                ids=ids, documents=docs, embeddings=embeddings, metadatas=metadatas
            )
            all_ids.extend(ids)
        return all_ids

    def _query_embedding(
        self,
        embedding: List[float],
        limit: int = 10,
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Perform vector search using pre-computed embedding.

        Args:
            embedding: Pre-computed embedding vector for semantic search
            limit: Maximum number of results to return (default: 10)
            collection_name: The collection to apply the query. Optional.
            **kwargs: ChromaDB-specific parameters (where, etc.)

        Returns:
            List of documents matching the search criteria
        """
        # Extract ChromaDB-specific parameters
        collection_name = collection_name or self.collection
        collection_metadata = kwargs.get(
            "collection_metadata", self.collection_metadata
        )
        create_collection_if_not_exists = kwargs.get(
            "create_collection_if_not_exists", self.create_collection_if_not_exists
        )
        where = kwargs.get("where")  # Metadata filters

        # Get or create collection based on configuration
        if create_collection_if_not_exists:
            # ChromaDB doesn't accept empty metadata, pass None instead
            metadata = collection_metadata if collection_metadata else None
            collection = self.client.get_or_create_collection(
                name=collection_name,
                metadata=metadata,
            )
        else:
            collection = self.client.get_collection(name=collection_name)

        # Perform query
        results = collection.query(
            query_embeddings=[embedding],
            n_results=limit,
            where=where,
            include=["documents", "metadatas"],
        )

        # Convert to Document objects
        documents = []
        if results["documents"] and results["documents"][0]:
            for i, doc_content in enumerate(results["documents"][0]):
                doc_id = results["ids"][0][i] if results["ids"] else None
                metadata = (
                    results["metadatas"][0][i]
                    if results["metadatas"] and results["metadatas"][0]
                    else {}
                )

                documents.append(
                    Document(
                        content=doc_content,
                        id=doc_id,
                        metadata=metadata,
                    )
                )

        return documents



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import json
import logging
from typing import TYPE_CHECKING

from flink_agents.api.agents.agent import Agent
from flink_agents.api.chat_message import ChatMessage, MessageRole
from flink_agents.api.decorators import (
    action,
    chat_model_setup,
    prompt,
)
from flink_agents.api.events.chat_event import ChatRequestEvent, ChatResponseEvent
from flink_agents.api.events.event import InputEvent, OutputEvent
from flink_agents.api.prompts.prompt import Prompt
from flink_agents.api.resource import ResourceDescriptor, ResourceName
from flink_agents.api.runner_context import RunnerContext
from flink_agents.examples.quickstart.agents.custom_types_and_resources import (
    ProductSuggestion,
    product_suggestion_prompt,
)

if TYPE_CHECKING:
    from flink_agents.examples.quickstart.agents.custom_types_and_resources import (
        ProductReviewSummary,
    )


class ProductSuggestionAgent(Agent):
    """An agent that uses a large language model (LLM) to generate actionable product
    improvement suggestions from aggregated product review data.

    This agent receives a summary of product reviews, including a rating distribution
    and a list of user dissatisfaction reasons, and produces concrete suggestions for
    product enhancement. It handles prompt construction, LLM interaction, and output
    parsing.
    """

    @prompt
    @staticmethod
    def generate_suggestion_prompt() -> Prompt:
        """Generate product suggestions based on the rating distribution and user
        dissatisfaction reasons.
        """
        return product_suggestion_prompt

    @chat_model_setup
    @staticmethod
    def generate_suggestion_model() -> ResourceDescriptor:
        """ChatModel which focus on generating product suggestions."""
        return ResourceDescriptor(
            clazz=ResourceName.ChatModel.OLLAMA_SETUP,
            connection="ollama_server",
            model="qwen3:8b",
            prompt="generate_suggestion_prompt",
            extract_reasoning=True,
        )

    @action(InputEvent)
    @staticmethod
    def process_input(event: InputEvent, ctx: RunnerContext) -> None:
        """Process input event."""
        input: ProductReviewSummary = event.input
        ctx.short_term_memory.set("id", input.id)
        ctx.short_term_memory.set("score_hist", input.score_hist)

        content = f"""
            "id": {input.id},
            "score_histogram": {input.score_hist},
            "unsatisfied_reasons": {input.unsatisfied_reasons}
        """
        ctx.send_event(
            ChatRequestEvent(
                model="generate_suggestion_model",
                messages=[
                    ChatMessage(role=MessageRole.USER, extra_args={"input": content})
                ],
            )
        )

    @action(ChatResponseEvent)
    @staticmethod
    def process_chat_response(event: ChatResponseEvent, ctx: RunnerContext) -> None:
        """Process chat response event."""
        try:
            json_content = json.loads(event.response.content)
            ctx.send_event(
                OutputEvent(
                    output=ProductSuggestion(
                        id=ctx.short_term_memory.get("id"),
                        score_hist=ctx.short_term_memory.get("score_hist"),
                        suggestions=json_content["suggestion_list"],
                    )
                )
            )
        except Exception:
            logging.exception(
                f"Error processing chat response {event.response.content}"
            )

            # To fail the agent, you can raise an exception here.



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List

from pydantic import BaseModel, Field
from typing_extensions import override

from flink_agents.api.resource import Resource, ResourceType


class VectorStoreQueryMode(str, Enum):
    """Vector store query mode.

    Attributes:
    ----------
    SEMANTIC : str
        Perform semantic search using embedding vectors to find
        the most semantically similar documents in the vector store.
    """

    SEMANTIC = "semantic"  # embedding-based retrieval
    # KEYWORD = "keyword"  # TODO: term-based retrieval
    # HYBRID = "hybrid"  # TODO: semantic + keyword retrieval


class VectorStoreQuery(BaseModel):
    """Structured query object for vector store operations.

    Provides a unified interface for text-based semantic search queries
    while supporting vector store-specific parameters and configurations.

    Attributes:
    ----------
    mode : VectorStoreQueryMode
        The type of query operation to perform (default: SEMANTIC).
    query_text : str
        Text query to be converted to embedding for semantic search.
    limit : int
        Maximum number of results to return (default: 10).
    collection_name : str
        The collection to apply the query. Optional.
    extra_args : Dict[str, Any]
        Vector store-specific parameters such as filters, distance metrics,
        namespaces, or other search configurations.
    """

    mode: VectorStoreQueryMode = Field(
        default=VectorStoreQueryMode.SEMANTIC,
        description="The type of query operation to perform.",
    )
    query_text: str = Field(
        description="Text query to be converted to embedding for semantic search."
    )
    limit: int = Field(default=10, description="Maximum number of results to return.")
    collection_name: str | None = Field(
        default=None, description="The collection to apply the query."
    )
    extra_args: Dict[str, Any] = Field(
        default_factory=dict, description="Vector store-specific parameters."
    )

    def __str__(self) -> str:
        return f"{self.mode.value} query: '{self.query_text}' (limit={self.limit})"


class Document(BaseModel):
    """A document retrieved from vector store search.

    Represents a single piece of content with associated metadata.

    Attributes:
    ----------
    content : str
        The actual text content of the document.
    metadata : Dict[str, Any]
        Document metadata such as source, author, timestamp, etc.
    id : Optional[str]
        Unique identifier of the document (if available).
    """

    content: str = Field(description="The actual text content of the document.")
    metadata: Dict[str, Any] = Field(
        default_factory=dict,
        description="Document metadata such as source, author, timestamp, etc.",
    )
    id: str | None = Field(
        default=None, description="Unique identifier of the document."
    )
    embedding: List[float] | None = Field(
        default=None, description="Embedding vector for content."
    )

    def __str__(self) -> str:
        content_preview = (
            self.content[:50] + "..." if len(self.content) > 50 else self.content
        )
        return f"Document: {content_preview}"


class VectorStoreQueryResult(BaseModel):
    """Result from a vector store query operation.

    Contains the retrieved documents from the search.

    Attributes:
    ----------
    documents : List[Document]
        List of documents retrieved from the vector store.
    """

    documents: List[Document] = Field(
        description="List of documents retrieved from the vector store."
    )

    def __str__(self) -> str:
        return f"QueryResult: {len(self.documents)} documents"


class BaseVectorStore(Resource, ABC):
    """Base abstract class for vector store.

    Provides vector store functionality that integrates embedding models
    for text-based semantic search. Handles both connection management and
    embedding generation internally.
    """

    embedding_model: str = Field(
        description="Name of the embedding model resource to use."
    )

    @classmethod
    @override
    def resource_type(cls) -> ResourceType:
        """Return resource type of class."""
        return ResourceType.VECTOR_STORE

    @property
    @abstractmethod
    def store_kwargs(self) -> Dict[str, Any]:
        """Return vector store setup settings passed to connection.

        These parameters are merged with query-specific parameters
        when performing vector search operations.
        """

    def add(
        self,
        documents: Document | List[Document],
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> List[str]:
        """Add documents to the vector store.

        Converts document content to embeddings and stores them in the vector store.
        The implementation may generate IDs for documents that don't have one.

        Args:
            documents: Single Document or list of Documents to add.
            collection_name: The collection name of the documents to add to. Optional.
            **kwargs: Vector store specific parameters.

        Returns:
            List of document IDs that were added to the vector store
        """
        # Normalize to list
        documents = _maybe_cast_to_list(documents)

        # Generate embeddings for all documents
        embedding_model = self.get_resource(
            self.embedding_model, ResourceType.EMBEDDING_MODEL
        )

        # Generate embeddings for each document
        for doc in documents:
            if doc.embedding is None:
                doc.embedding = embedding_model.embed(doc.content)

        # Merge setup kwargs with add-specific args
        merged_kwargs = self.store_kwargs.copy()
        merged_kwargs.update(kwargs)

        # Perform add operation using the abstract method
        return self._add_embedding(
            documents=documents, collection_name=collection_name, **merged_kwargs
        )

    def query(self, query: VectorStoreQuery) -> VectorStoreQueryResult:
        """Perform vector search using structured query object.

        Converts text query to embeddings and returns structured query result.

        Args:
            query: VectorStoreQuery object containing query parameters

        Returns:
            VectorStoreQueryResult containing the retrieved documents
        """
        # Generate embedding from the query text
        embedding_model = self.get_resource(
            self.embedding_model, ResourceType.EMBEDDING_MODEL
        )
        query_embedding = embedding_model.embed(query.query_text)

        # Merge setup kwargs with query-specific args
        merged_kwargs = self.store_kwargs.copy()
        merged_kwargs.update(query.extra_args)

        # Perform vector search using the abstract method
        documents = self._query_embedding(
            query_embedding, query.limit, query.collection_name, **merged_kwargs
        )

        # Return structured result
        return VectorStoreQueryResult(documents=documents)

    @abstractmethod
    def size(self, collection_name: str | None = None) -> int:
        """Get the size of the collection in vector store.

        Args:
            collection_name: The target collection. Optional.
        """

    @abstractmethod
    def get(
        self,
        ids: str | List[str] | None = None,
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Retrieve a document from the vector store by its ID.

        Args:
            ids: Unique identifier of the documents to retrieve
            collection_name: The collection name of the documents to retrieve. Optional.
            **kwargs: Vector store specific parameters (offset, limit, filter etc.)

        Returns:
            Document object if found, None otherwise
        """

    @abstractmethod
    def delete(
        self,
        ids: str | List[str] | None = None,
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Delete documents in the vector store by its IDs.

        Args:
            ids: Unique identifier of the documents to delete
            collection_name: The collection name of the documents belong to. Optional.
            **kwargs: Vector store specific parameters (filter etc.)
        """

    @abstractmethod
    def _query_embedding(
        self,
        embedding: List[float],
        limit: int = 10,
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> List[Document]:
        """Perform vector search using pre-computed embedding.

        Args:
            embedding: Pre-computed embedding vector for semantic search
            limit: Maximum number of results to return (default: 10)
            collection_name: The collection to apply the query. Optional.
            **kwargs: Vector store-specific parameters (filters, distance metrics, etc.)

        Returns:
            List of documents matching the search criteria
        """

    @abstractmethod
    def _add_embedding(
        self,
        *,
        documents: List[Document],
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> List[str]:
        """Add documents with pre-computed embeddings to the vector store.

        Args:
            documents: Documents with embeddings to add to the vector store
            collection_name: The collection name of the documents to add. Optional.
            **kwargs: Vector store-specific parameters (collection, namespace, etc.)

        Returns:
            List of document IDs that were added to the vector store
        """


class Collection(BaseModel):
    """Represents a collection of documents."""

    name: str
    metadata: Dict[str, Any] | None = None


class CollectionManageableVectorStore(BaseVectorStore, ABC):
    """Base abstract class for vector store which support collection management."""

    @abstractmethod
    def get_or_create_collection(
        self, name: str, metadata: Dict[str, Any] | None = None
    ) -> Collection:
        """Get a collection, or create it if it doesn't exist.

        Args:
            name: Name of the collection
            metadata: Metadata of the collection
        Returns:
            The retrieved or created collection
        """

    @abstractmethod
    def get_collection(self, name: str) -> Collection:
        """Get a collection, raise an exception if it doesn't exist.

        Args:
            name: Name of the collection
        Returns:
            The retrieved collection
        """

    @abstractmethod
    def delete_collection(self, name: str) -> Collection:
        """Delete a collection.

        Args:
            name: Name of the collection
        Returns:
            The deleted collection
        """


def _maybe_cast_to_list(value: Any | List[Any]) -> List[Any] | None:
    """Cast T to List[T] if T is not None."""
    if value is None:
        return None
    return [value] if not isinstance(value, list) else value



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import hashlib
from abc import ABC
from typing import Any, Dict

try:
    from typing import override
except ImportError:
    from typing_extensions import override
from uuid import UUID

from pydantic import BaseModel, Field, model_validator
from pydantic_core import PydanticSerializationError
from pyflink.common import Row


class Event(BaseModel, ABC, extra="allow"):
    """Base class for all event types in the system.

    Event allows extra properties, but these must be BaseModel instances or JSON
    serializable.

    Attributes:
    ----------
    id : UUID
        Unique identifier for the event, generated deterministically based on
        event content. This ensures events with identical content have the same
        ID, which is critical for ActionStateStore divergence detection.
    """

    id: UUID = Field(default=None)

    @staticmethod
    def __serialize_unknown(field: Any) -> Dict[str, Any]:
        """Handle serialization of unknown types, specifically Row objects."""
        if isinstance(field, Row):
            return {"type": "Row", "values": field._values}
        else:
            err_msg = f"Unable to serialize unknown type: {field.__class__}"
            raise PydanticSerializationError(err_msg)

    @override
    def model_dump_json(self, **kwargs: Any) -> str:
        """Override model_dump_json to handle Row objects using fallback."""
        # Set fallback if not provided in kwargs
        if "fallback" not in kwargs:
            kwargs["fallback"] = self.__serialize_unknown
        return super().model_dump_json(**kwargs)

    def _generate_content_based_id(self) -> UUID:
        """Generate a deterministic UUID based on event content using MD5 hash.

        Similar to Java's UUID.nameUUIDFromBytes(), uses MD5 for version 3 UUID.
        """
        # Serialize content excluding 'id' to avoid circular dependency
        content_json = super().model_dump_json(
            exclude={"id"}, fallback=self.__serialize_unknown
        )
        md5_hash = hashlib.md5(content_json.encode()).digest()
        return UUID(bytes=md5_hash, version=3)

    @model_validator(mode="after")
    def validate_and_set_id(self) -> "Event":
        """Validate that fields are serializable and generate content-based ID."""
        if self.id is None:
            object.__setattr__(self, "id", self._generate_content_based_id())
        self.model_dump_json()
        return self

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        # Ensure added property can be serialized.
        self.model_dump_json()
        # Regenerate ID if content changed (but not if setting 'id' itself)
        if name != "id":
            object.__setattr__(self, "id", self._generate_content_based_id())


class InputEvent(Event):
    """Event generated by the framework, carrying an input data that
    arrives at the agent.
    """

    input: Any


class OutputEvent(Event):
    """Event representing a result from agent. By generating an OutputEvent,
    actions can emit output data.

    Attributes:
    ----------
    output : Any
        The output result returned by the agent.
    """

    output: Any



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import os
import sysconfig
import tempfile
from datetime import datetime, timezone
from importlib import resources
from pathlib import Path
from typing import Any, List

from pydantic import BaseModel
from pyflink.common import Encoder, Types, WatermarkStrategy
from pyflink.datastream import (
    KeySelector,
    RuntimeExecutionMode,
    StreamExecutionEnvironment,
)
from pyflink.datastream.connectors.file_system import (
    FileSource,
    StreamFormat,
    StreamingFileSink,
)

from flink_agents.api.agents.agent import Agent
from flink_agents.api.core_options import AgentConfigOptions
from flink_agents.api.decorators import (
    action,
    chat_model_connection,
    chat_model_setup,
    embedding_model_connection,
    embedding_model_setup,
    vector_store,
)
from flink_agents.api.events.event import Event, InputEvent, OutputEvent
from flink_agents.api.execution_environment import AgentsExecutionEnvironment
from flink_agents.api.memory.long_term_memory import (
    CompactionConfig,
    LongTermMemoryBackend,
    LongTermMemoryOptions,
    MemorySetItem,
)
from flink_agents.api.resource import (
    ResourceDescriptor,
    ResourceName,
)
from flink_agents.api.runner_context import RunnerContext
from flink_agents.e2e_tests.test_utils import pull_model
from flink_agents.integrations.vector_stores.chroma.chroma_vector_store import (
    ChromaVectorStore,
)

current_dir = Path(__file__).parent

os.environ["PYTHONPATH"] = sysconfig.get_paths()["purelib"]

chromadb_path = tempfile.mkdtemp()

OLLAMA_CHAT_MODEL = "qwen3:8b"
OLLAMA_EMBEDDING_MODEL = "nomic-embed-text"
pull_model(OLLAMA_CHAT_MODEL)
pull_model(OLLAMA_EMBEDDING_MODEL)


class ItemData(BaseModel):
    """Data model for storing item information.

    Attributes:
    ----------
    id : int
        Unique identifier of the item
    review : str
        The user review of the item
    review_score: float
        The review_score of the item
    """

    id: int
    review: str
    review_score: float
    memory_info: dict | None = None


class Record(BaseModel):  # noqa: D101
    id: int
    count: int
    timestamp_before_add: str
    timestamp_after_add: str
    timestamp_second_action: str | None = None
    items: List[MemorySetItem] | None = None


class MyEvent(Event):  # noqa D101
    value: Any


class MyKeySelector(KeySelector):
    """KeySelector for extracting key."""

    def get_key(self, value: ItemData) -> int:
        """Extract key from ItemData."""
        return value.id


class LongTermMemoryAgent(Agent):
    """Agent used for testing long term memory ."""

    @chat_model_connection
    @staticmethod
    def ollama_connection() -> ResourceDescriptor:
        """ChatModelConnection responsible for ollama model service connection."""
        return ResourceDescriptor(
            clazz=ResourceName.ChatModel.OLLAMA_CONNECTION, request_timeout=480.0
        )

    @chat_model_setup
    @staticmethod
    def ollama_qwen3() -> ResourceDescriptor:
        """ChatModel which focus on math, and reuse ChatModelConnection."""
        return ResourceDescriptor(
            clazz=ResourceName.ChatModel.OLLAMA_SETUP,
            connection="ollama_connection",
            model=OLLAMA_CHAT_MODEL,
            extract_reasoning=True,
            think=False,
        )

    @embedding_model_connection
    @staticmethod
    def ollama_embedding_connection() -> ResourceDescriptor:  # noqa D102
        return ResourceDescriptor(
            clazz=ResourceName.EmbeddingModel.OLLAMA_CONNECTION, request_timeout=240.0
        )

    @embedding_model_setup
    @staticmethod
    def ollama_nomic_embed_text() -> ResourceDescriptor:  # noqa D102
        return ResourceDescriptor(
            clazz=ResourceName.EmbeddingModel.OLLAMA_SETUP,
            connection="ollama_embedding_connection",
            model=OLLAMA_EMBEDDING_MODEL,
        )

    @vector_store
    @staticmethod
    def chroma_vector_store() -> ResourceDescriptor:
        """Vector store setup for knowledge base."""
        return ResourceDescriptor(
            clazz=ResourceName.VectorStore.CHROMA_VECTOR_STORE,
            embedding_model="ollama_nomic_embed_text",
            persist_directory=chromadb_path,
        )

    @action(InputEvent)
    @staticmethod
    async def add_items(event: Event, ctx: RunnerContext):  # noqa D102
        input_data = event.input
        ltm = ctx.long_term_memory

        timestamp_before_add = datetime.now(timezone.utc).isoformat()
        memory_set = ltm.get_or_create_memory_set(
            name="test_ltm",
            item_type=str,
            capacity=5,
            compaction_config=CompactionConfig(model="ollama_qwen3"),
        )
        await ctx.durable_execute_async(memory_set.add, items=input_data.review)
        timestamp_after_add = datetime.now(timezone.utc).isoformat()

        stm = ctx.short_term_memory
        count = stm.get("count") or 1
        stm.set("count", count + 1)

        ctx.send_event(
            MyEvent(
                value=Record(
                    id=input_data.id,
                    count=count,
                    timestamp_before_add=timestamp_before_add,
                    timestamp_after_add=timestamp_after_add,
                )
            )
        )

    @action(MyEvent)
    @staticmethod
    async def retrieve_items(event: Event, ctx: RunnerContext):  # noqa D102
        record: Record = event.value
        record.timestamp_second_action = datetime.now(timezone.utc).isoformat()
        memory_set = ctx.long_term_memory.get_memory_set(name="test_ltm")
        items = await ctx.durable_execute_async(memory_set.get)
        if (
            (record.id == 1 and record.count == 3)
            or (record.id == 2 and record.count == 5)
            or (record.id == 3 and record.count == 2)
        ):
            record.items = items
        ctx.send_event(OutputEvent(output=record))


def test_long_term_memory_async_execution_in_action(tmp_path: Path) -> None:  # noqa: D103
    env = StreamExecutionEnvironment.get_execution_environment()
    env.set_runtime_mode(RuntimeExecutionMode.STREAMING)
    env.set_parallelism(1)

    # currently, bounded source is not supported due to runtime implementation, so
    # we use continuous file source here.
    input_datastream = env.from_source(
        source=FileSource.for_record_stream_format(
            StreamFormat.text_line_format(), f"file:///{current_dir}/../resources/input"
        ).build(),
        watermark_strategy=WatermarkStrategy.no_watermarks(),
        source_name="streaming_agent_example",
    )

    deserialize_datastream = input_datastream.map(
        lambda x: ItemData.model_validate_json(x)
    )

    agents_env = AgentsExecutionEnvironment.get_execution_environment(env=env)
    agents_config = agents_env.get_config()
    agents_config.set(AgentConfigOptions.JOB_IDENTIFIER, "LTM_TEST_JOB")
    agents_config.set(
        LongTermMemoryOptions.BACKEND, LongTermMemoryBackend.EXTERNAL_VECTOR_STORE
    )
    agents_config.set(
        LongTermMemoryOptions.EXTERNAL_VECTOR_STORE_NAME, "chroma_vector_store"
    )
    agents_config.set(LongTermMemoryOptions.ASYNC_COMPACTION, True)

    output_datastream = (
        agents_env.from_datastream(
            input=deserialize_datastream, key_selector=MyKeySelector()
        )
        .apply(LongTermMemoryAgent())
        .to_datastream()
    )

    result_dir = tmp_path / "results"
    result_dir.mkdir(parents=True, exist_ok=True)

    output_datastream.map(lambda x: x.model_dump_json(), Types.STRING()).add_sink(
        StreamingFileSink.for_row_format(
            base_path=str(result_dir.absolute()),
            encoder=Encoder.simple_string_encoder(),
        ).build()
    )

    agents_env.execute()

    check_result(result_dir=result_dir)


def check_result(*, result_dir: Path) -> None:  # noqa: D103
    actual_result = []
    for file in result_dir.iterdir():
        if file.is_dir():
            for child in file.iterdir():
                with child.open() as f:
                    actual_result.extend(
                        [Record.model_validate_json(line) for line in f]
                    )

    records = {}
    for record in actual_result:
        records[f"{record.id}.{record.count}"] = record

    # verify async add doesn't block process other key
    assert datetime.fromisoformat(
        records["2.1"].timestamp_before_add
    ) < datetime.fromisoformat(records["1.1"].timestamp_after_add)
    assert datetime.fromisoformat(
        records["3.1"].timestamp_before_add
    ) < datetime.fromisoformat(records["1.1"].timestamp_after_add)

    # verify async compaction doesn't block any operation
    assert not records["2.5"].items[0].compacted
    store = ChromaVectorStore(
        persist_directory=chromadb_path, embedding_model="ollama_nomic_embed_text"
    )
    doc = store.get(collection_name="LTM_TEST_JOB--89360337-test_ltm")
    print(f"Retrieved items: {doc}")
    log_dir = resources.files("pyflink").joinpath("log")

    log = ""
    for item in log_dir.iterdir():
        if item.is_file() and item.name.endswith(".log"):
            log = item.read_text()
    assert len(doc) == 1, log
    doc = doc[0]
    assert doc.metadata.get("compacted")



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import importlib
import json
import typing
from typing import Any, Callable, Dict

import cloudpickle

from flink_agents.api.chat_message import ChatMessage, MessageRole
from flink_agents.api.events.event import Event, InputEvent
from flink_agents.api.resource import Resource, ResourceType, get_resource_class
from flink_agents.api.tools.tool import Tool, ToolMetadata
from flink_agents.api.tools.utils import (
    create_java_tool_schema_str_from_model,
    create_model_from_java_tool_schema_str,
)
from flink_agents.api.vector_stores.vector_store import (
    Collection,
    Document,
    VectorStoreQuery,
    VectorStoreQueryMode,
    VectorStoreQueryResult,
)
from flink_agents.plan.resource_provider import JAVA_RESOURCE_MAPPING
from flink_agents.runtime.java.java_resource_wrapper import (
    JavaGetResourceWrapper,
    JavaPrompt,
    JavaTool,
)


def convert_to_python_object(bytesObject: bytes) -> Any:
    """Used for deserializing Python objects."""
    return cloudpickle.loads(bytesObject)


def _build_event_log_string(event: InputEvent | Event, event_type: str) -> str:
    try:
        payload = json.loads(event.model_dump_json())
        payload["eventType"] = event_type
        payload.setdefault("attributes", {})
        return json.dumps(payload)
    except Exception:
        return str(event)


def wrap_to_input_event(bytesObject: bytes) -> tuple[bytes, str]:
    """Wrap data to python input event and serialize.

    Returns:
        A tuple of (serialized_event_bytes, event_json_str)
    """
    event = InputEvent(input=cloudpickle.loads(bytesObject))
    event_type = f"{event.__class__.__module__}.{event.__class__.__qualname__}"
    return (cloudpickle.dumps(event), _build_event_log_string(event, event_type))


def get_output_from_output_event(bytesObject: bytes) -> Any:
    """Get output data from OutputEvent and serialize."""
    return cloudpickle.dumps(convert_to_python_object(bytesObject).output)

def create_resource(resource_module: str, resource_clazz: str, func_kwargs: Dict[str, Any]) -> Resource:
    """Dynamically create a resource instance from module and class name.

    Args:
        resource_module: The module path containing the resource class
        resource_clazz: The class name to instantiate
        func_kwargs: Keyword arguments to pass to the class constructor

    Returns:
        Resource: An instance of the specified resource class
    """
    module = importlib.import_module(resource_module)
    cls = getattr(module, resource_clazz)
    return cls(**func_kwargs)

def get_resource_function(j_resource_adapter: Any) -> Callable:
    """Create a callable wrapper for Java resource adapter.

    Args:
        j_resource_adapter: Java resource adapter object

    Returns:
        Callable: A Python callable that wraps the Java resource adapter
    """
    return JavaGetResourceWrapper(j_resource_adapter).get_resource

def from_java_tool(j_tool: Any) -> JavaTool:
    """Convert a Java tool object to a Python JavaTool instance.

    Args:
        j_tool: Java tool object

    Returns:
        JavaTool: Python wrapper for the Java tool with extracted metadata
    """
    name = j_tool.getName()
    metadata = ToolMetadata(
        name=name,
        description=j_tool.getDescription(),
        args_schema=create_model_from_java_tool_schema_str(name, j_tool.getMetadata().getInputSchema()),
    )
    return JavaTool(metadata=metadata)

def from_java_prompt(j_prompt: Any) -> JavaPrompt:
    """Convert a Java prompt object to a Python JavaPrompt instance.

    Args:
        j_prompt: Java prompt object to be wrapped

    Returns:
        JavaPrompt: Python wrapper for the Java prompt
    """
    return JavaPrompt(j_prompt=j_prompt)

def from_java_resource(type_name: str, kwargs: Dict[str, Any]) -> Resource:
    """Convert a Java resource object to a Python Resource instance.
    This function is used to convert a Java resource object to a Python Resource
    instance.

    Args:
        type_name: Java resource type name
        kwargs: Keyword arguments
    Returns:
        Resource: Python wrapper for the Java resource
    """
    class_path = JAVA_RESOURCE_MAPPING.get(ResourceType(type_name))
    if not class_path:
        err_msg = f"No Java resource mapping found for {type_name}"
        raise ValueError(err_msg)

    module_path, class_name = class_path.rsplit(".", 1)
    cls = get_resource_class(module_path, class_name)

    return cls(**kwargs)

def normalize_tool_call_id(tool_call: Dict[str, Any]) -> Dict[str, Any]:
    """Normalize tool call by converting the ID field to string format while preserving
    all other fields.

    This function ensures that the tool call ID is consistently represented as a string,
    which is required for compatibility with certain systems that expect string IDs.

    Args:
        tool_call: Dictionary containing tool call information. The dictionary may
                   contain any number of fields, but typically includes:
                  - id: Tool call identifier (will be converted to string)
                  - type: Tool call type (preserved as-is)
                  - function: Function details (preserved as-is)
                  - Any other fields (preserved as-is)
    """
    normalized_call = tool_call.copy()

    normalized_call["id"] = str(tool_call.get("id", ""))

    return normalized_call

def from_java_chat_message(j_chat_message: Any) -> ChatMessage:
    """Convert a chat message to a python chat message."""
    return ChatMessage(role=MessageRole(j_chat_message.getRole().getValue()),
                       content=j_chat_message.getContent(),
                       tool_calls=[normalize_tool_call_id(tool_call) for tool_call in j_chat_message.getToolCalls()],
                       extra_args=j_chat_message.getExtraArgs())


def to_java_chat_message(chat_message: ChatMessage) -> Any:
    """Convert a chat message to a java chat message."""
    from pemja import findClass
    j_ChatMessage = findClass("org.apache.flink.agents.api.chat.messages.ChatMessage")
    j_chat_message = j_ChatMessage()

    j_MessageRole = findClass("org.apache.flink.agents.api.chat.messages.MessageRole")
    j_chat_message.setRole(j_MessageRole.fromValue(chat_message.role.value))
    j_chat_message.setContent(chat_message.content)
    j_chat_message.setExtraArgs(chat_message.extra_args)
    if chat_message.tool_calls:
        tool_calls = [normalize_tool_call_id(tool_call) for tool_call in chat_message.tool_calls]
        j_chat_message.setToolCalls(tool_calls)

    return j_chat_message

# TODO: Replace this with `to_java_chat_message()` when the `find_class` bug is fixed.
def update_java_chat_message(chat_message: ChatMessage, j_chat_message: Any) -> str:
    """Update a Java chat message using Python chat message."""
    j_chat_message.setContent(chat_message.content)
    j_chat_message.setExtraArgs(chat_message.extra_args)
    if chat_message.tool_calls:
        tool_calls = [normalize_tool_call_id(tool_call) for tool_call in chat_message.tool_calls]
        j_chat_message.setToolCalls(tool_calls)

    return chat_message.role.value

def from_java_document(j_document: Any) -> Document:
    """Convert a Java documents to a Python document."""
    document = Document(
        content=j_document.getContent(),
        id=j_document.getId(),
        metadata=j_document.getMetadata(),
    )
    if j_document.getEmbedding():
        document.embedding = list(j_document.getEmbedding())
    return document

def update_java_document(document: Document, j_document: Any) -> None:
    """Update a Java document using Python document."""
    j_document.setContent(document.content)
    j_document.setId(document.id)
    j_document.setMetadata(document.metadata)
    if document.embedding:
        j_document.setEmbedding(tuple(document.embedding))


def from_java_vector_store_query(j_query: Any) -> VectorStoreQuery:
    """Convert a Java vector store query to a Python query."""
    return VectorStoreQuery(
        mode=VectorStoreQueryMode(j_query.getMode().getValue()),
        query_text=j_query.getQueryText(),
        limit=j_query.getLimit(),
        collection_name=j_query.getCollection(),
        extra_args=j_query.getExtraArgs()
    )

def from_java_vector_store_query_result(j_query: Any) -> VectorStoreQueryResult:
    """Convert a Java vector store query result to a Python query result."""
    return VectorStoreQueryResult(
        documents=[from_java_document(j_document) for j_document in j_query.getDocuments()],
    )

def from_java_collection(j_collection: Any) -> Collection:
    """Convert a Java collection to a Python collection."""
    return Collection(
        name=j_collection.getName(),
        metadata=j_collection.getMetadata(),
    )

def from_java_message_role(j_role: Any) -> MessageRole:
    """Convert a Java message role to a Python message role."""
    return MessageRole(j_role.getValue())

def get_java_tool_metadata_from_tool(tool: Tool) -> typing.Dict[str, str]:
    """Retrieve Java format tool metadata from a tool input schema string."""
    return {"name": tool.name, "description": tool.metadata.description, "inputSchema": create_java_tool_schema_str_from_model(tool.metadata.args_schema)}

def get_mode_value(query: VectorStoreQuery) -> str:
    """Get the mode value of a VectorStoreQuery."""
    return query.mode.value

def call_method(obj: Any, method_name: str, kwargs: Dict[str, Any]) -> Any:
    """Calls a method on `obj` by name and passes in positional and keyword arguments.

    Parameters:
        obj: Any Python object
        method_name: A string representing the name of the method to call
        kwargs: Keyword arguments to pass to the method

    Returns:
        The return value of the method

    Raises:
        AttributeError: If the object does not have the specified method
    """
    if not hasattr(obj, method_name):
        err_msg = f"Object {obj} has no attribute '{method_name}'"
        raise AttributeError(err_msg)

    method = getattr(obj, method_name)
    return method(**kwargs)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
from typing import List

import pytest

from flink_agents.api.decorators import action
from flink_agents.api.events.event import Event, InputEvent, OutputEvent
from flink_agents.api.runner_context import RunnerContext


def test_action_decorator() -> None:  # noqa D103
    @action(InputEvent)
    def forward_action(event: Event, ctx: RunnerContext) -> None:
        input = event.input
        ctx.send_event(OutputEvent(output=input))

    assert hasattr(forward_action, "_listen_events")
    listen_events = forward_action._listen_events
    assert listen_events == (InputEvent,)


def test_action_decorator_listen_multi_events() -> None:  # noqa D103
    @action(InputEvent, OutputEvent)
    def forward_action(event: Event, ctx: RunnerContext) -> None:
        input = event.input
        ctx.send_event(OutputEvent(output=input))

    assert hasattr(forward_action, "_listen_events")
    listen_events = forward_action._listen_events
    assert listen_events == (InputEvent, OutputEvent)


def test_action_decorator_listen_no_event() -> None:  # noqa D103
    with pytest.raises(AssertionError):

        @action()
        def forward_action(event: Event, ctx: RunnerContext) -> None:
            input = event.input
            ctx.send_event(OutputEvent(output=input))


def test_action_decorator_listen_non_event_type() -> None:  # noqa D103
    with pytest.raises(AssertionError):

        @action(List)
        def forward_action(event: Event, ctx: RunnerContext) -> None:
            input = event.input
            ctx.send_event(OutputEvent(output=input))



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import functools
import queue
import uuid
from concurrent.futures import Future
from datetime import datetime, timezone
from typing import Any, Dict, List, Type, cast

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import override

from flink_agents.api.chat_message import ChatMessage
from flink_agents.api.memory.long_term_memory import (
    CompactionConfig,
    DatetimeRange,
    ItemType,
    LongTermMemoryOptions,
    MemorySet,
    MemorySetItem,
)
from flink_agents.api.metric_group import MetricGroup
from flink_agents.api.resource import ResourceType
from flink_agents.api.runner_context import RunnerContext
from flink_agents.api.vector_stores.vector_store import (
    CollectionManageableVectorStore,
    Document,
    VectorStoreQuery,
    _maybe_cast_to_list,
)
from flink_agents.runtime.memory.compaction_functions import summarize
from flink_agents.runtime.memory.internal_base_long_term_memory import (
    InternalBaseLongTermMemory,
)


# TODO: support async execution for operations and compaction
class VectorStoreLongTermMemory(InternalBaseLongTermMemory):
    """Long-Term Memory based on ChromaDB."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    ctx: RunnerContext = Field(
        description="The runner context to retrieve resource.", exclude=True
    )

    vector_store: str | CollectionManageableVectorStore = Field(
        description="The vector store backend to store data."
    )

    job_id: str = Field(description="Unique identifier for the job.")

    key: str = Field(
        default=None, description="Unique identifier for the keyed partition."
    )

    async_compaction: bool = Field(
        default=False, description="Whether to execute compact asynchronously."
    )

    metric_group: MetricGroup = Field(
        default=None, description="Metric group for reporting long-term memory metrics."
    )
    metric_records: queue.Queue = Field(
        default=queue.Queue(), description="A thread safe queue for record metrics."
    )

    def __init__(
        self,
        *,
        ctx: RunnerContext,
        vector_store: str,
        job_id: str,
        **kwargs: Any,
    ) -> None:
        """Init method."""
        super().__init__(
            ctx=ctx,
            vector_store=vector_store,
            job_id=job_id,
            async_compaction=ctx.config.get(LongTermMemoryOptions.ASYNC_COMPACTION),
            metric_group=ctx.agent_metric_group.get_sub_group("long-term-memory"),
            **kwargs,
        )

    @override
    def switch_context(self, key: str) -> None:
        self.key = key

    @property
    def store(self) -> CollectionManageableVectorStore:
        """Get backend vector store.

        The vector store used for long-term memory must implement
        CollectionManageableVectorStore.
        """
        if isinstance(self.vector_store, str):
            self.vector_store = cast(
                "CollectionManageableVectorStore",
                self.ctx.get_resource(self.vector_store, ResourceType.VECTOR_STORE),
            )

        return self.vector_store

    @override
    def get_or_create_memory_set(
        self,
        name: str,
        item_type: type[str] | Type[ChatMessage],
        capacity: int,
        compaction_config: CompactionConfig,
    ) -> MemorySet:
        memory_set = MemorySet(
            name=name,
            item_type=item_type,
            capacity=capacity,
            compaction_config=compaction_config,
            ltm=self,
        )
        self.store.get_or_create_collection(
            name=self._name_mangling(name),
            metadata={"memory_set": memory_set.model_dump_json()},
        )
        return memory_set

    @override
    def get_memory_set(self, name: str) -> MemorySet:
        collection = self.store.get_collection(name=self._name_mangling(name))
        memory_set = MemorySet.model_validate_json(collection.metadata["memory_set"])
        memory_set.ltm = self
        return memory_set

    @override
    def delete_memory_set(self, name: str) -> bool:
        return self.store.delete_collection(name=self._name_mangling(name)) is not None

    @override
    def size(self, memory_set: MemorySet) -> int:
        return self.store.size(collection_name=self._name_mangling(memory_set.name))

    @override
    def add(
        self,
        memory_set: MemorySet,
        memory_items: ItemType | List[ItemType],
        ids: str | List[str] | None = None,
        metadatas: Dict[str, Any] | List[Dict[str, Any]] | None = None,
    ) -> List[str]:
        memory_items = _maybe_cast_to_list(memory_items)
        ids = _maybe_cast_to_list(ids)
        metadatas = _maybe_cast_to_list(metadatas)

        if ids is None:
            ids = [str(uuid.uuid4()) for _ in memory_items]

        timestamp = datetime.now(timezone.utc).isoformat()
        meta = {
            "compacted": False,
            "created_time": timestamp,
            "last_accessed_time": timestamp,
        }

        merge_metadatas = [meta for _ in memory_items]

        if metadatas is not None:
            for merge_meta, metadata in zip(merge_metadatas, metadatas, strict=False):
                merge_meta.update(metadata)

        if issubclass(memory_set.item_type, BaseModel):
            memory_items = [item.model_dump_json() for item in memory_items]

        documents = []
        for id, item, metadata in zip(ids, memory_items, merge_metadatas, strict=False):
            documents.append(
                Document(
                    id=id,
                    content=item,
                    metadata=metadata,
                )
            )

        ids = self.store.add(
            documents=documents, collection_name=self._name_mangling(memory_set.name)
        )

        if memory_set.size >= memory_set.capacity:
            # trigger compaction
            if self.async_compaction:
                future = self.ctx.executor.submit(
                    self._compact,
                    memory_set=memory_set,
                    metric_group=self.metric_group,
                )
                future.add_done_callback(
                    functools.partial(
                        self._handle_exception, self.job_id, self.key, memory_set
                    )
                )
            else:
                self._compact(
                    memory_set=memory_set,
                    metric_group=self.metric_group,
                )

        self._report_token_metrics()
        return ids

    @override
    def get(
        self, memory_set: MemorySet, ids: str | List[str] | None = None
    ) -> List[MemorySetItem]:
        documents = self.store.get(
            ids=ids, collection_name=self._name_mangling(memory_set.name)
        )
        return self._convert_to_items(memory_set=memory_set, documents=documents)

    @override
    def delete(self, memory_set: MemorySet, ids: str | List[str] | None = None) -> None:
        self.store.delete(ids=ids, collection_name=self._name_mangling(memory_set.name))

    @override
    def search(
        self, memory_set: MemorySet, query: str, limit: int, **kwargs: Any
    ) -> List[MemorySetItem]:
        query = VectorStoreQuery(
            query_text=query,
            limit=limit,
            collection_name=self._name_mangling(memory_set.name),
            extra_args=kwargs,
        )
        result = self.store.query(query=query)

        return self._convert_to_items(memory_set=memory_set, documents=result.documents)

    @override
    def close(self) -> None:
        # report possible token usage metrics
        self._report_token_metrics()

    def _report_token_metrics(self) -> None:
        """Report token usage metrics."""
        if not self.metric_records.empty():
            if self.metric_group is None:
                return
            while not self.metric_records.empty():
                metric = self.metric_records.get()
                if (
                    metric.get("model_name")
                    and metric.get("promptTokens")
                    and metric.get("completionTokens")
                ):
                    model_group = self.metric_group.get_sub_group(metric["model_name"])
                    model_group.get_counter("promptTokens").inc(metric["promptTokens"])
                    model_group.get_counter("completionTokens").inc(
                        metric["completionTokens"]
                    )

    def _name_mangling(self, name: str) -> str:
        """Mangle memory set name to actually name in vector store."""
        return f"{self.job_id}-{self.key}-{name}"

    def _compact(self, memory_set: MemorySet, metric_group: MetricGroup) -> Any | None:
        """Compact memory set to manage storge."""
        # currently, only support summarize all the items.
        extra_args = summarize(
            ltm=self, memory_set=memory_set, ctx=self.ctx, metric_group=metric_group
        )
        self.metric_records.put(extra_args)

    @staticmethod
    def _handle_exception(
        job_id: str, key: str, memory_set: MemorySet, future: Future
    ) -> None:
        exception = future.exception()
        if exception is not None:
            err_msg = f"Compaction for {job_id}-{key}-{memory_set.name} failed."
            # TODO: Currently, this exception will appear in the log of TaskManager,
            # but will not cause the Flink job to fail.
            raise RuntimeError(err_msg) from exception

    @staticmethod
    def _convert_to_items(
        memory_set: MemorySet,
        documents: List[Document],
    ) -> List[MemorySetItem]:
        """Convert retrival documents to memory items."""
        items = []
        for doc in documents:
            metadata = doc.metadata
            item = MemorySetItem(
                memory_set_name=memory_set.name,
                id=doc.id,
                value=memory_set.item_type.model_validate_json(doc.content)
                if issubclass(memory_set.item_type, BaseModel)
                else doc.content,
                compacted=metadata.pop("compacted"),
                last_accessed_time=datetime.fromisoformat(
                    metadata["last_accessed_time"]
                ),
            )
            if item.compacted:
                item.created_time = DatetimeRange(
                    start=datetime.fromisoformat(metadata.pop("created_time_start")),
                    end=datetime.fromisoformat(metadata.pop("created_time_end")),
                )
            else:
                item.created_time = datetime.fromisoformat(metadata.pop("created_time"))

            if len(metadata) > 0:
                item.additional_metadata = metadata

            items.append(item)
        return items



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import subprocess
from pathlib import Path

from ollama import Client

current_dir = Path(__file__).parent


def pull_model(ollama_model: str) -> Client:
    """Run ollama pull ollama_model."""
    try:
        # prepare ollama server
        subprocess.run(
            ["bash", f"{current_dir}/scripts/ollama_pull_model.sh", ollama_model],
            timeout=120,
            check=True,
        )
        client = Client()
        models = client.list()

        model_found = False
        for model in models["models"]:
            if model.model == ollama_model:
                model_found = True
                break

        if not model_found:
            client = None  # type: ignore
    except Exception:
        client = None  # type: ignore

    return client


def check_result(*, result_dir: Path, groud_truth_dir: Path) -> None:
    """Util function for checking flink job execution result."""
    actual_result = []
    for file in result_dir.iterdir():
        if file.is_dir():
            for child in file.iterdir():
                with child.open() as f:
                    actual_result.extend(f.readlines())
        if file.is_file():
            with file.open() as f:
                actual_result.extend(f.readlines())

    with groud_truth_dir.open() as f:
        expected = f.readlines().sort()

    actual_result = actual_result.sort()
    assert actual_result == expected



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
from typing import Any, Callable

from docstring_parser import parse
from typing_extensions import override

from flink_agents.api.tools.tool import Tool, ToolMetadata, ToolType
from flink_agents.api.tools.utils import create_schema_from_function
from flink_agents.plan.function import JavaFunction, PythonFunction


class FunctionTool(Tool):
    """Tool that takes in a function.

    Attributes:
    ----------
    func : Function
        User defined function.
    """

    func: PythonFunction | JavaFunction

    @classmethod
    @override
    def tool_type(cls) -> ToolType:
        """Get the tool type."""
        return ToolType.FUNCTION

    def call(self, *args: Any, **kwargs: Any) -> Any:
        """Call the function tool."""
        return self.func(*args, **kwargs)


def from_callable(func: Callable) -> FunctionTool:
    """Create FunctionTool from a user defined function.

    Parameters
    ----------
    func : Callable
        The function to analyze.
    """
    description = parse(func.__doc__).description
    metadata = ToolMetadata(
        name=func.__name__,
        description=description,
        args_schema=create_schema_from_function(func.__name__, func=func),
    )
    return FunctionTool(func=PythonFunction.from_callable(func), metadata=metadata)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import os
import time

import pytest

from flink_agents.api.agents.agent import Agent
from flink_agents.api.decorators import (
    action,
    embedding_model_connection,
    embedding_model_setup,
    vector_store,
)
from flink_agents.api.events.context_retrieval_event import (
    ContextRetrievalRequestEvent,
    ContextRetrievalResponseEvent,
)
from flink_agents.api.events.event import InputEvent, OutputEvent
from flink_agents.api.resource import (
    ResourceDescriptor,
    ResourceName,
    ResourceType,
)
from flink_agents.api.runner_context import RunnerContext
from flink_agents.api.vector_stores.vector_store import (
    CollectionManageableVectorStore,
    Document,
)

TEST_COLLECTION = "test_collection"
MAX_RETRIES_TIMES = 10

class VectorStoreCrossLanguageAgent(Agent):
    """Example agent demonstrating cross-language embedding model testing."""

    @embedding_model_connection
    @staticmethod
    def embedding_model_connection() -> ResourceDescriptor:
        """EmbeddingModelConnection responsible for ollama model service connection."""
        if os.environ.get("EMBEDDING_TYPE") == "JAVA":
            return ResourceDescriptor(
                clazz=ResourceName.EmbeddingModel.JAVA_WRAPPER_CONNECTION,
                java_clazz=ResourceName.EmbeddingModel.Java.OLLAMA_CONNECTION,
                host="http://localhost:11434",
            )
        return ResourceDescriptor(
            clazz=ResourceName.EmbeddingModel.OLLAMA_CONNECTION,
            host="http://localhost:11434",
        )

    @embedding_model_setup
    @staticmethod
    def embedding_model() -> ResourceDescriptor:
        """EmbeddingModel which focus on math, and reuse ChatModelConnection."""
        if os.environ.get("EMBEDDING_TYPE") == "JAVA":
            return ResourceDescriptor(
                clazz=ResourceName.EmbeddingModel.JAVA_WRAPPER_SETUP,
                java_clazz=ResourceName.EmbeddingModel.Java.OLLAMA_SETUP,
                connection="embedding_model_connection",
                model=os.environ.get(
                    "OLLAMA_EMBEDDING_MODEL", "nomic-embed-text:latest"
                ),
            )
        return ResourceDescriptor(
            clazz=ResourceName.EmbeddingModel.OLLAMA_SETUP,
            connection="embedding_model_connection",
            model=os.environ.get("OLLAMA_EMBEDDING_MODEL", "nomic-embed-text:latest"),
        )

    @vector_store
    @staticmethod
    def vector_store() -> ResourceDescriptor:
        """Vector store setup for knowledge base."""
        return ResourceDescriptor(
            clazz=ResourceName.VectorStore.JAVA_WRAPPER_COLLECTION_MANAGEABLE_VECTOR_STORE,
            java_clazz=ResourceName.VectorStore.Java.ELASTICSEARCH_VECTOR_STORE,
            embedding_model="embedding_model",
            host=os.environ.get("ES_HOST"),
            index="my_documents",
            dims=768,
        )

    @action(InputEvent)
    @staticmethod
    def process_input(event: InputEvent, ctx: RunnerContext) -> None:
        """User defined action for processing input.

        In this action, we will test Vector store Collection Management and
        Document Management.
        """
        input_text = event.input

        stm = ctx.short_term_memory

        is_initialized = stm.get("is_initialized") or False

        if not is_initialized:
            print("[TEST] Initializing vector store...")

            vector_store = ctx.get_resource("vector_store", ResourceType.VECTOR_STORE)
            if isinstance(vector_store, CollectionManageableVectorStore):
                vector_store.get_or_create_collection(TEST_COLLECTION , metadata={"key1": "value1", "key2": "value2"})

                collection = vector_store.get_collection(name=TEST_COLLECTION)

                assert collection is not None
                assert collection.name == TEST_COLLECTION
                assert collection.metadata == {"key1": "value1", "key2": "value2"}

                vector_store.delete_collection(name=TEST_COLLECTION)

                with pytest.raises(RuntimeError):
                    vector_store.get_collection(name=TEST_COLLECTION)

                print("[TEST] Vector store Collection Management PASSED")

                documents = [
                    Document(
                        id="doc1",
                        content="The sum of 1 and 2 equals 3.",
                        metadata={"category": "calculate", "source": "test"},
                    ),
                    Document(
                        id="doc2",
                        content="Why did the cat sit on the computer? Because it wanted to keep an eye on the mouse.",
                        metadata={"category": "ai-agent", "source": "test"},
                    ),
                    Document(
                        id="doc3",
                        content="This is a test document used to verify the delete functionality.",
                        metadata={"category": "utility", "source": "test"},
                    ),
                ]
                vector_store.add(documents=documents)

                # Test size
                assert vector_store.size() == 3

                # Test delete
                vector_store.delete(ids="doc3")

                # Wait for vector store to delete doc3
                retry_time = 0
                while vector_store.size() > 2 and retry_time < MAX_RETRIES_TIMES:
                    retry_time += 1
                    time.sleep(2)
                    print(f"[TEST] Retrying to delete doc3, retry_time={retry_time}")

                assert vector_store.size() == 2

                # Test get
                doc = vector_store.get(ids="doc2")
                assert doc is not None
                assert doc[0].id == "doc2"
                assert doc[0].content == "Why did the cat sit on the computer? Because it wanted to keep an eye on the mouse."

                print("[TEST] Vector store Document Management PASSED")

            stm.set("is_initialized", True)


        ctx.send_event(ContextRetrievalRequestEvent(query=input_text, vector_store="vector_store"))

    @action(ContextRetrievalResponseEvent)
    @staticmethod
    def contextRetrievalResponseEvent(event: ContextRetrievalResponseEvent, ctx: RunnerContext) -> None:
        """User defined action for processing context retrieval response.

        In this action, we will test Vector store Context Retrieval.
        """
        documents = event.documents

        assert documents is not None
        assert len(documents) > 0

        for document in documents:
            assert document is not None
            assert document.id is not None
            assert document.content is not None

        test_result = f"[PASS] retrieved_count={len(documents)}, first_doc_id={documents[0].id}, first_doc_preview={documents[0].content[:50]}"
        print(f"[TEST] Vector store Context Retrieval PASSED, first_doc_id={documents[0].id}, first_doc_preview={documents[0].content[:50]}")

        ctx.send_event(OutputEvent(output=test_result))



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import os
import subprocess
import sys
from pathlib import Path

import pytest
from ollama import Client

from flink_agents.api.resource import Resource, ResourceType
from flink_agents.integrations.embedding_models.local.ollama_embedding_model import (
    OllamaEmbeddingModelConnection,
    OllamaEmbeddingModelSetup,
)

test_model = os.environ.get("OLLAMA_EMBEDDING_MODEL", "all-minilm:22m")
current_dir = Path(__file__).parent

try:
    # only auto setup ollama in ci with python 3.10 to reduce ci cost.
    if "3.10" in sys.version:
        subprocess.run(
            ["bash", f"{current_dir}/start_ollama_server.sh"], timeout=300, check=True
        )
    client = Client()
    models = client.list()

    model_found = False
    for model in models["models"]:
        if model.model == test_model:
            model_found = True
            break

    if not model_found:
        client = None  # type: ignore
except Exception:
    client = None  # type: ignore


@pytest.mark.skipif(
    client is None, reason="Ollama client is not available or test embedding model is missing"
)
def test_ollama_embedding_setup() -> None:
    """Test embedding functionality with OllamaEmbeddingModelSetup."""
    connection = OllamaEmbeddingModelConnection(
        name="ollama_embed",
        base_url="http://localhost:11434"
    )

    def get_resource(name: str, type: ResourceType) -> Resource:
        return connection

    setup = OllamaEmbeddingModelSetup(
        name="embeddings",
        connection="ollama_embed",
        model=test_model,
        truncate=True,
        get_resource=get_resource
    )

    # Test embedding through setup
    embedding = setup.embed("This is a test sentence for embedding.")
    assert embedding is not None
    assert isinstance(embedding, list)
    assert len(embedding) > 0
    assert all(isinstance(x, float) for x in embedding)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import json
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Type, cast

from flink_agents.api.chat_message import ChatMessage, MessageRole
from flink_agents.api.memory.long_term_memory import (
    BaseLongTermMemory,
    CompactionConfig,
    MemorySet,
    MemorySetItem,
)
from flink_agents.api.metric_group import MetricGroup
from flink_agents.api.prompts.prompt import Prompt
from flink_agents.api.resource import ResourceType
from flink_agents.api.runner_context import RunnerContext

if TYPE_CHECKING:
    from flink_agents.api.chat_models.chat_model import BaseChatModelSetup


DEFAULT_ANALYSIS_PROMPT = Prompt.from_text("""<role>
Context Summarize Assistant
</role>

<primary_objective>
Your sole objective in this task is to summarize the context above.
</primary_objective>

<objective_information>
You're nearing the total number of input tokens you can accept, so you need compact the context. To achieve this objective, you should extract important topics. Notice,
**The topics must no more than {limit}**. Afterwards, you should generate summarization for each topic, and and record which messages the summary was derived from.
The message index start from 0.
</objective_information>

<output_example>
You must always respond with valid json format in this format:
{"topic1": {"summarization": "User ask what is 1 * 2, and the result is 3.", "messages": [0,1,2]},
 ...
 "topic4": {"summarization": "User ask what's the weather tomorrow, llm use the search_weather, and the answer is snow.", "messages": [9,10,11,12]}
}
</output_example>
""")


def summarize(
    ltm: BaseLongTermMemory,
    memory_set: MemorySet,
    ctx: RunnerContext,
    metric_group: MetricGroup,
    ids: List[str] | None = None,
) -> Dict[str, Any]:
    """Generate summarization of the items in the memory set.

    Will add the summarization to memory set, and delete original items involved
    in summarization.

    Args:
        ltm: The long term memory the memory set belongs to.
        memory_set: The memory set to be summarized.
        ctx: The runner context used to retrieve needed resources.
        metric_group: Metric group used to report metrics.
        ids: The ids of items to be summarized. If not provided, all items will be
        involved in summarization. Optional
    """
    compaction_config: CompactionConfig = memory_set.compaction_config

    # retrieve all items
    items: List[MemorySetItem] = ltm.get(memory_set=memory_set, ids=ids)

    response: ChatMessage = _generate_summarization(
        items, memory_set.item_type, compaction_config, ctx, metric_group
    )

    logging.debug(f"Items to be summarized: {items}\nSummarization: {response.content}")

    for topic in cast("dict", json.loads(response.content)).values():
        summarization = topic["summarization"]
        indices = topic["messages"]

        if compaction_config.limit == 1:
            indices = list(range(len(items)))

        if memory_set.item_type == ChatMessage:
            item = ChatMessage(role=MessageRole.USER, content=summarization)
        else:
            item = summarization

        create_time_list = []
        for index in indices:
            if items[index].compacted:
                create_time_list.append(items[index].created_time.start)
                create_time_list.append(items[index].created_time.end)
            else:
                create_time_list.append(items[index].created_time)

        start = min(create_time_list).isoformat()
        end = max(create_time_list).isoformat()
        last_accessed_time = max(
            [items[index].last_accessed_time for index in indices]
        ).isoformat()

        # delete involved items
        ids = [items[index].id for index in indices]
        ltm.delete(memory_set=memory_set, ids=ids)

        # add summarization to memory set
        ltm.add(
            memory_set=memory_set,
            memory_items=item,
            metadatas={
                "compacted": True,
                "created_time_start": start,
                "created_time_end": end,
                "last_accessed_time": last_accessed_time,
            },
        )

    return response.extra_args


# TODO: Currently, we feed all items to the LLM at once, which may exceed the LLM's
# context window. We need to support batched summary generation.
def _generate_summarization(
    memory_set_items: List[MemorySetItem],
    item_type: Type,
    compaction_config: CompactionConfig,
    ctx: RunnerContext,
    metric_group: MetricGroup
) -> ChatMessage:
    """Generate summarization of the items by llm."""
    # get arguments
    model_name = compaction_config.model
    prompt = compaction_config.prompt

    msgs: List[ChatMessage]
    if item_type == ChatMessage:
        msgs = [item.value for item in memory_set_items]
    else:
        msgs = [
            ChatMessage(role=MessageRole.USER, content=str(item.value))
            for item in memory_set_items
        ]

    # generate summary
    model: BaseChatModelSetup = cast(
        "BaseChatModelSetup",
        ctx.get_resource(name=model_name, type=ResourceType.CHAT_MODEL, metric_group=metric_group),
    )
    input_variable = {}
    for msg in msgs:
        input_variable.update(msg.extra_args)

    if prompt is not None:
        if isinstance(prompt, str):
            prompt: Prompt = cast(
                "Prompt",
                ctx.get_resource(prompt, ResourceType.PROMPT, metric_group=metric_group),
            )
        prompt_messages = prompt.format_messages(
            role=MessageRole.USER, **input_variable
        )
        msgs.extend(prompt_messages)
    else:
        msgs.extend(DEFAULT_ANALYSIS_PROMPT.format_messages(limit=str(compaction_config.limit)))

    response: ChatMessage = model.chat(messages=msgs)

    return response



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import json
from pathlib import Path
from typing import Any, Dict, List, Sequence

import pytest

from flink_agents.api.agents.agent import Agent
from flink_agents.api.chat_message import ChatMessage, MessageRole
from flink_agents.api.chat_models.chat_model import BaseChatModelSetup
from flink_agents.api.decorators import (
    action,
    chat_model_setup,
    embedding_model_connection,
    embedding_model_setup,
    vector_store,
)
from flink_agents.api.embedding_models.embedding_model import (
    BaseEmbeddingModelConnection,
    BaseEmbeddingModelSetup,
)
from flink_agents.api.events.event import Event, InputEvent, OutputEvent
from flink_agents.api.resource import ResourceDescriptor, ResourceType
from flink_agents.api.runner_context import RunnerContext
from flink_agents.api.vector_stores.vector_store import (
    BaseVectorStore,
    Document,
)
from flink_agents.plan.agent_plan import AgentPlan
from flink_agents.plan.configuration import AgentConfiguration
from flink_agents.plan.function import PythonFunction


class AgentForTest(Agent):  # noqa D101
    @action(InputEvent)
    @staticmethod
    def increment(event: Event, ctx: RunnerContext) -> None:  # noqa D102
        value = event.input
        value += 1
        ctx.send_event(OutputEvent(output=value))


def test_from_agent():  # noqa D102
    agent = AgentForTest()
    agent_plan = AgentPlan.from_agent(agent, AgentConfiguration())
    event_type = f"{InputEvent.__module__}.{InputEvent.__name__}"
    actions = agent_plan.get_actions(event_type)
    assert len(actions) == 1
    action = actions[0]
    assert action.name == "increment"
    func = action.exec
    assert isinstance(func, PythonFunction)
    assert func.module == "flink_agents.plan.tests.test_agent_plan"
    assert func.qualname == "AgentForTest.increment"
    assert action.listen_event_types == [event_type]


class InvalidAgent(Agent):  # noqa D101
    @action(InputEvent)
    @staticmethod
    def invalid_signature_action(event: Event) -> None:  # noqa D102
        pass


def test_to_agent_invalid_signature() -> None:  # noqa D103
    agent = InvalidAgent()
    with pytest.raises(TypeError):
        AgentPlan.from_agent(agent, AgentConfiguration())


class MyEvent(Event):
    """Event for testing purposes."""


class MockChatModelImpl(BaseChatModelSetup):  # noqa: D101
    host: str
    desc: str

    @property
    def model_kwargs(self) -> Dict[str, Any]:  # noqa: D102
        return {}

    @classmethod
    def resource_type(cls) -> ResourceType:  # noqa: D102
        return ResourceType.CHAT_MODEL

    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatMessage:
        """Testing Implementation."""
        return ChatMessage(
            role=MessageRole.ASSISTANT, content=self.host + " " + self.desc
        )


class MockEmbeddingModelConnection(BaseEmbeddingModelConnection):  # noqa: D101
    api_key: str

    def embed(self, text: str | Sequence[str], **kwargs: Any) -> list[float]:
        """Testing Implementation."""
        if isinstance(text, str):
            return [0.1234, -0.5678, 0.9012, -0.3456, 0.7890]
        return [[0.1234, -0.5678, 0.9012, -0.3456, 0.7890]]


class MockEmbeddingModelSetup(BaseEmbeddingModelSetup):  # noqa: D101
    @property
    def model_kwargs(self) -> Dict[str, Any]:  # noqa: D102
        return {"model": self.model}


class MockVectorStore(BaseVectorStore):  # noqa: D101
    host: str
    port: int
    collection_name: str

    @property
    def store_kwargs(self) -> Dict[str, Any]:  # noqa: D102
        return {"collection_name": self.collection_name}

    def size(self, collection_name: str | None = None) -> int:
        """For Testing."""

    def get(
        self,
        ids: str | List[str] | None = None,
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> List[Document]:
        """For Testing."""

    def delete(
        self,
        ids: str | List[str] | None = None,
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> None:
        """For Testing."""

    def _add_embedding(
        self,
        *,
        documents: List[Document],
        collection_name: str | None = None,
        **kwargs: Any,
    ) -> List[str]:
        """For Testing."""

    def _query_embedding(
        self, embedding: list[float], limit: int = 10, **kwargs: Any
    ) -> list[Document]:
        """Testing Implementation."""
        return [
            Document(
                content="Mock document content",
                metadata={"source": "test", "id": "doc1"},
                id="doc1",
            ),
            Document(
                content="Another mock document",
                metadata={"source": "test", "id": "doc2"},
                id="doc2",
            ),
        ][:limit]


class MyAgent(Agent):  # noqa: D101
    @chat_model_setup
    @staticmethod
    def mock() -> ResourceDescriptor:  # noqa: D102
        return ResourceDescriptor(
            clazz=f"{MockChatModelImpl.__module__}.{MockChatModelImpl.__name__}",
            host="8.8.8.8",
            desc="mock resource just for testing.",
            connection="mock",
        )

    @embedding_model_connection
    @staticmethod
    def mock_embedding_conn() -> ResourceDescriptor:  # noqa: D102
        return ResourceDescriptor(
            clazz=f"{MockEmbeddingModelConnection.__module__}.{MockEmbeddingModelConnection.__name__}", api_key="mock-api-key"
        )

    @embedding_model_setup
    @staticmethod
    def mock_embedding() -> ResourceDescriptor:  # noqa: D102
        return ResourceDescriptor(
            clazz=f"{MockEmbeddingModelSetup.__module__}.{MockEmbeddingModelSetup.__name__}",
            model="test-model",
            connection="mock_embedding_conn",
        )

    @vector_store
    @staticmethod
    def mock_vector_store() -> ResourceDescriptor:  # noqa: D102
        return ResourceDescriptor(
            clazz=f"{MockVectorStore.__module__}.{MockVectorStore.__name__}",
            embedding_model="mock_embedding",
            host="localhost",
            port=8000,
            collection_name="test_collection",
        )

    @action(InputEvent)
    @staticmethod
    def first_action(event: InputEvent, ctx: RunnerContext) -> None:  # noqa: D102
        pass

    @action(InputEvent, MyEvent)
    @staticmethod
    def second_action(event: InputEvent, ctx: RunnerContext) -> None:  # noqa: D102
        pass


@pytest.fixture(scope="module")
def agent_plan() -> AgentPlan:  # noqa: D103
    return AgentPlan.from_agent(
        MyAgent(), AgentConfiguration({"mock.key": "mock.value"})
    )


current_dir = Path(__file__).parent


def test_agent_plan_serialize(agent_plan: AgentPlan) -> None:  # noqa: D103
    json_value = agent_plan.model_dump_json(serialize_as_any=True, indent=4)
    with Path.open(Path(f"{current_dir}/resources/agent_plan.json")) as f:
        expected_json = f.read()
    actual = json.loads(json_value)
    expected = json.loads(expected_json)
    assert actual == expected


def test_agent_plan_deserialize(agent_plan: AgentPlan) -> None:  # noqa: D103
    with Path.open(Path(f"{current_dir}/resources/agent_plan.json")) as f:
        expected_json = f.read()
    deserialized_agent_plan = AgentPlan.model_validate_json(expected_json)
    assert deserialized_agent_plan == agent_plan


def test_get_resource() -> None:  # noqa: D103
    agent_plan = AgentPlan.from_agent(MyAgent(), AgentConfiguration())
    mock = agent_plan.get_resource("mock", ResourceType.CHAT_MODEL)
    assert (
        mock.chat(ChatMessage(role=MessageRole.USER, content="")).content
        == "8.8.8.8 mock resource just for testing."
    )


def test_add_action_and_resource_to_agent() -> None:  # noqa: D103
    my_agent = Agent()
    my_agent.add_action(
        name="first_action", events=[InputEvent], func=MyAgent.first_action
    )
    my_agent.add_action(
        name="second_action", events=[InputEvent, MyEvent], func=MyAgent.second_action
    )
    my_agent.add_resource(
        name="mock",
        resource_type=ResourceType.CHAT_MODEL,
        instance=ResourceDescriptor(
            clazz=f"{MockChatModelImpl.__module__}.{MockChatModelImpl.__name__}",
            host="8.8.8.8",
            desc="mock resource just for testing.",
            connection="mock",
        ),
    )

    my_agent.add_resource(
        name="mock_embedding_conn",
        resource_type=ResourceType.EMBEDDING_MODEL_CONNECTION,
        instance=ResourceDescriptor(
            clazz=f"{MockEmbeddingModelConnection.__module__}.{MockEmbeddingModelConnection.__name__}", api_key="mock-api-key"
        ),
    )
    my_agent.add_resource(
        name="mock_embedding",
        resource_type=ResourceType.EMBEDDING_MODEL,
        instance=ResourceDescriptor(
            clazz=f"{MockEmbeddingModelSetup.__module__}.{MockEmbeddingModelSetup.__name__}",
            model="test-model",
            connection="mock_embedding_conn",
        ),
    )
    my_agent.add_resource(
        name="mock_vector_store",
        resource_type=ResourceType.VECTOR_STORE,
        instance=ResourceDescriptor(
            clazz=f"{MockVectorStore.__module__}.{MockVectorStore.__name__}",
            embedding_model="mock_embedding",
            host="localhost",
            port=8000,
            collection_name="test_collection",
        ),
    )
    agent_plan = AgentPlan.from_agent(
        my_agent, AgentConfiguration({"mock.key": "mock.value"})
    )
    json_value = agent_plan.model_dump_json(serialize_as_any=True, indent=4)
    with Path.open(Path(f"{current_dir}/resources/agent_plan.json")) as f:
        expected_json = f.read()
    actual = json.loads(json_value)
    expected = json.loads(expected_json)
    assert actual == expected



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
from typing import TYPE_CHECKING, Any, Dict, List, cast

from pydantic import BaseModel, field_serializer, model_validator

from flink_agents.api.agents.agent import Agent
from flink_agents.api.resource import Resource, ResourceDescriptor, ResourceType
from flink_agents.plan.actions.action import Action
from flink_agents.plan.actions.chat_model_action import CHAT_MODEL_ACTION
from flink_agents.plan.actions.context_retrieval_action import CONTEXT_RETRIEVAL_ACTION
from flink_agents.plan.actions.tool_call_action import TOOL_CALL_ACTION
from flink_agents.plan.configuration import AgentConfiguration
from flink_agents.plan.function import PythonFunction
from flink_agents.plan.resource_provider import (
    JavaResourceProvider,
    JavaSerializableResourceProvider,
    PythonResourceProvider,
    PythonSerializableResourceProvider,
    ResourceProvider,
)
from flink_agents.plan.tools.function_tool import from_callable

if TYPE_CHECKING:
    from flink_agents.integrations.mcp.mcp import MCPServer

BUILT_IN_ACTIONS = [CHAT_MODEL_ACTION, TOOL_CALL_ACTION, CONTEXT_RETRIEVAL_ACTION]


class AgentPlan(BaseModel):
    """Agent plan compiled from user defined agent.

    Attributes:
    ----------
    actions: Dict[str, Action]
        Mapping of action names to actions
    actions_by_event : Dict[Type[Event], str]
        Mapping of event types to the list of actions name that listen to them.
    resource_providers: ResourceProvider
        Two level mapping of resource type to resource name to resource provider.
    """

    actions: Dict[str, Action]
    actions_by_event: Dict[str, List[str]]
    resource_providers: Dict[ResourceType, Dict[str, ResourceProvider]] | None = None
    config: AgentConfiguration | None = None
    __resources: Dict[ResourceType, Dict[str, Resource]] = {}
    __j_resource_adapter: Any = None

    @field_serializer("resource_providers")
    def __serialize_resource_providers(
        self, providers: Dict[ResourceType, Dict[str, ResourceProvider]]
    ) -> dict:
        # append meta info to help deserialize resource providers
        data = {}
        for type in providers:
            data[type] = {}
            for name, provider in providers[type].items():
                data[type][name] = provider.model_dump()
                if isinstance(provider, PythonResourceProvider):
                    data[type][name]["__resource_provider_type__"] = (
                        "PythonResourceProvider"
                    )
                elif isinstance(provider, PythonSerializableResourceProvider):
                    data[type][name]["__resource_provider_type__"] = (
                        "PythonSerializableResourceProvider"
                    )
                elif isinstance(provider, JavaResourceProvider):
                    data[type][name]["__resource_provider_type__"] = (
                        "JavaResourceProvider"
                    )
                elif isinstance(provider, JavaSerializableResourceProvider):
                    data[type][name]["__resource_provider_type__"] = (
                        "JavaSerializableResourceProvider"
                    )
        return data

    @model_validator(mode="before")
    def __custom_deserialize(self) -> "AgentPlan":
        if "resource_providers" in self:
            providers = self["resource_providers"]
            # restore exec from serialized json.
            if isinstance(providers, dict):
                for type in providers:
                    for name, provider in providers[type].items():
                        if isinstance(provider, dict):
                            provider_type = provider["__resource_provider_type__"]
                            if provider_type == "PythonResourceProvider":
                                self["resource_providers"][type][name] = (
                                    PythonResourceProvider.model_validate(provider)
                                )
                            elif provider_type == "PythonSerializableResourceProvider":
                                self["resource_providers"][type][name] = (
                                    PythonSerializableResourceProvider.model_validate(
                                        provider
                                    )
                                )
                            elif provider_type == "JavaResourceProvider":
                                self["resource_providers"][type][name] = (
                                    JavaResourceProvider.model_validate(provider)
                                )
                            elif provider_type == "JavaSerializableResourceProvider":
                                self["resource_providers"][type][name] = (
                                    JavaSerializableResourceProvider.model_validate(
                                        provider
                                    )
                                )
        return self

    @staticmethod
    def from_agent(agent: Agent, config: AgentConfiguration) -> "AgentPlan":
        """Build a AgentPlan from user defined agent."""
        actions = {}
        actions_by_event = {}
        for action in _get_actions(agent) + BUILT_IN_ACTIONS:
            assert action.name not in actions, f"Duplicate action name: {action.name}"
            actions[action.name] = action
            for event_type in action.listen_event_types:
                if event_type not in actions_by_event:
                    actions_by_event[event_type] = []
                actions_by_event[event_type].append(action.name)

        resource_providers = {}
        for provider in _get_resource_providers(agent, config):
            type = provider.type
            if type not in resource_providers:
                resource_providers[type] = {}
            name = provider.name
            assert name not in resource_providers[type], (
                f"Duplicate resource name: {name}"
            )
            resource_providers[type][name] = provider
        return AgentPlan(
            actions=actions,
            actions_by_event=actions_by_event,
            resource_providers=resource_providers,
            config=config,
        )

    def get_actions(self, event_type: str) -> List[Action]:
        """Get actions that listen to the specified event type.

        Parameters
        ----------
        event_type : Type[Event]
            The event type to query.

        Returns:
        -------
        list[Action]
            List of Actions that will respond to this event type.
        """
        return [self.actions[name] for name in self.actions_by_event[event_type]]

    def get_action_config(self, action_name: str) -> Dict[str, Any]:
        """Get config of the action.

        Parameters
        ----------
        action_name : str
            The name of the action.

        Returns:
        -------
        Dict[str, Any]
            The config of action.
        """
        return self.actions[action_name].config

    def get_action_config_value(self, action_name: str, key: str) -> Any:
        """Get config of the action.

        Parameters
        ----------
        action_name : str
            The name of the action.
        key : str
            The name of the option.

        Returns:
        -------
        Dict[str, Any]
            The option value of the action config.
        """
        return self.actions[action_name].config.get(key, None)

    def get_resource(self, name: str, type: ResourceType) -> Resource:
        """Get resource from agent plan.

        Parameters
        ----------
        name : str
            The name of the resource.
        type : ResourceType
            The type of the resource.
        """
        if type not in self.__resources:
            self.__resources[type] = {}
        if name not in self.__resources[type]:
            resource_provider = self.resource_providers[type][name]
            if isinstance(resource_provider, JavaResourceProvider):
                resource_provider.set_java_resource_adapter(self.__j_resource_adapter)
            resource = resource_provider.provide(
                get_resource=self.get_resource, config=self.config
            )
            self.__resources[type][name] = resource
        return self.__resources[type][name]

    def set_java_resource_adapter(self, j_resource_adapter: Any) -> None:
        """Set java resource adapter for java resource provider."""
        self.__j_resource_adapter = j_resource_adapter

    def close(self) -> None:
        """Clean up the resources."""
        for type in self.__resources:
            for name in self.__resources[type]:
                self.__resources[type][name].close()
        self.__resources.clear()



def _get_actions(agent: Agent) -> List[Action]:
    """Extract all registered agent actions from an agent.

    Parameters
    ----------
    agent : Agent
        The agent to be analyzed.

    Returns:
    -------
    List[Action]
        List of Action defined in the agent.
    """
    actions = []
    for name, value in agent.__class__.__dict__.items():
        if isinstance(value, staticmethod) and hasattr(value, "_listen_events"):
            actions.append(
                Action(
                    name=name,
                    exec=PythonFunction.from_callable(value.__func__),
                    listen_event_types=[
                        f"{event_type.__module__}.{event_type.__name__}"
                        for event_type in value._listen_events
                    ],
                )
            )
        elif callable(value) and hasattr(value, "_listen_events"):
            actions.append(
                Action(
                    name=name,
                    exec=PythonFunction.from_callable(value),
                    listen_event_types=[
                        f"{event_type.__module__}.{event_type.__name__}"
                        for event_type in value._listen_events
                    ],
                )
            )
    for name, action in agent.actions.items():
        actions.append(
            Action(
                name=name,
                exec=PythonFunction.from_callable(action[1]),
                listen_event_types=[
                    f"{event_type.__module__}.{event_type.__name__}"
                    for event_type in action[0]
                ],
                config=action[2],
            )
        )
    return actions


def _get_resource_providers(agent: Agent, config: AgentConfiguration) -> List[ResourceProvider]:
    resource_providers = []
    # retrieve resource declared by decorator
    for name, value in agent.__class__.__dict__.items():
        if (
            hasattr(value, "_is_chat_model_setup")
            or hasattr(value, "_is_chat_model_connection")
            or hasattr(value, "_is_embedding_model_setup")
            or hasattr(value, "_is_embedding_model_connection")
            or hasattr(value, "_is_vector_store")
        ):
            if isinstance(value, staticmethod):
                value = value.__func__

            if callable(value):
                descriptor = value()
                if hasattr(descriptor.clazz, "_is_java_resource"):
                    resource_providers.append(
                        JavaResourceProvider.get(name=name, descriptor=value())
                    )
                else:
                    resource_providers.append(
                        PythonResourceProvider.get(name=name, descriptor=value())
                    )

        elif hasattr(value, "_is_tool"):
            if isinstance(value, staticmethod):
                value = value.__func__

            if callable(value):
                # TODO: support other tool type.
                tool = from_callable(func=value)
                resource_providers.append(
                    PythonSerializableResourceProvider.from_resource(
                        name=name, resource=tool
                    )
                )
        elif hasattr(value, "_is_prompt"):
            if isinstance(value, staticmethod):
                value = value.__func__
            prompt = value()
            resource_providers.append(
                PythonSerializableResourceProvider.from_resource(
                    name=name, resource=prompt
                )
            )
        elif hasattr(value, "_is_mcp_server"):
            if isinstance(value, staticmethod):
                value = value.__func__

            descriptor = value()
            _add_mcp_server(name, resource_providers, descriptor, config)

    # retrieve resource declared by add interface
    for name, prompt in agent.resources[ResourceType.PROMPT].items():
        resource_providers.append(
            PythonSerializableResourceProvider.from_resource(name=name, resource=prompt)
        )

    for name, tool in agent.resources[ResourceType.TOOL].items():
        resource_providers.append(
            PythonSerializableResourceProvider.from_resource(
                name=name, resource=from_callable(tool.func)
            )
        )

    for name, descriptor in agent.resources[ResourceType.MCP_SERVER].items():
        _add_mcp_server(name, resource_providers, descriptor)

    for resource_type in [
        ResourceType.CHAT_MODEL,
        ResourceType.CHAT_MODEL_CONNECTION,
        ResourceType.EMBEDDING_MODEL,
        ResourceType.EMBEDDING_MODEL_CONNECTION,
        ResourceType.VECTOR_STORE,
    ]:
        for name, descriptor in agent.resources[resource_type].items():
            if hasattr(descriptor.clazz, "_is_java_resource"):
                resource_providers.append(
                    JavaResourceProvider.get(name=name, descriptor=descriptor)
                )
            else:
                resource_providers.append(
                    PythonResourceProvider.get(name=name, descriptor=descriptor)
                )

    return resource_providers


def _add_mcp_server(
    name: str, resource_providers: List[ResourceProvider], descriptor: ResourceDescriptor, config: AgentConfiguration
) -> None:
    provider = PythonResourceProvider.get(name=name, descriptor=descriptor)

    resource_providers.append(provider)

    def get_resource(name: str, descriptor: ResourceDescriptor) -> Any:
        """Placeholder."""

    mcp_server = cast("MCPServer", provider.provide(get_resource=get_resource, config=config))

    resource_providers.extend(
        [
            PythonSerializableResourceProvider.from_resource(
                name=prompt.name, resource=prompt
            )
            for prompt in mcp_server.list_prompts()
        ]
    )

    resource_providers.extend(
        [
            PythonSerializableResourceProvider.from_resource(
                name=tool.name, resource=tool
            )
            for tool in mcp_server.list_tools()
        ]
    )

    mcp_server.close()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
#################################################################################
import json
from pathlib import Path

import pytest

from flink_agents.api.resource import Resource, ResourceDescriptor, ResourceType
from flink_agents.plan.resource_provider import PythonResourceProvider, ResourceProvider

current_dir = Path(__file__).parent


class MockChatModelImpl(Resource):  # noqa: D101
    host: str
    desc: str

    @classmethod
    def resource_type(cls) -> ResourceType:  # noqa: D102
        return ResourceType.CHAT_MODEL


@pytest.fixture(scope="module")
def resource_provider() -> ResourceProvider:  # noqa: D103
    return PythonResourceProvider(
        name="mock",
        type=MockChatModelImpl.resource_type(),
        descriptor=ResourceDescriptor(clazz=f"{MockChatModelImpl.__module__}.{MockChatModelImpl.__name__}", host="8.8.8.8", desc="mock chat model"),
    )


def test_python_resource_provider_serialize(  # noqa: D103
    resource_provider: ResourceProvider,
) -> None:
    json_value = resource_provider.model_dump_json(serialize_as_any=True)
    with Path.open(Path(f"{current_dir}/resources/resource_provider.json")) as f:
        expected_json = f.read()
    actual = json.loads(json_value)
    expected = json.loads(expected_json)
    assert actual == expected


def test_python_resource_provider_deserialize(  # noqa: D103
    resource_provider: ResourceProvider,
) -> None:
    with Path.open(Path(f"{current_dir}/resources/resource_provider.json")) as f:
        expected_json = f.read()
    expected_resource_provider = PythonResourceProvider.model_validate_json(
        expected_json
    )
    assert resource_provider == expected_resource_provider
